"""
AF (and perhaps other arrhythmias like preamature beats) detection
using rr time series as input and using lstm as model.

References
----------
1. https://github.com/al3xsh/rnn-based-af-detection

"""

import warnings
from copy import deepcopy
from typing import Any, Optional, Sequence, Union, List

import torch
from torch import Tensor, nn
from einops import rearrange
from einops.layers.torch import Rearrange

from ..cfg import CFG
from ..components.outputs import BaseOutput
from ..model_configs.rr_lstm import RR_LSTM_CONFIG
from ..models._nets import (
    ExtendedCRF,
    GlobalContextBlock,
    NonLocalBlock,
    SEBlock,
    SelfAttention,
    MLP,
    StackedLSTM,
)
from ..utils.misc import CitationMixin
from ..utils.utils_nn import CkptMixin, SizeMixin


__all__ = [
    "RR_LSTM",
]


class RR_LSTM(nn.Module, CkptMixin, SizeMixin, CitationMixin):
    """LSTM model for RR time series classification or sequence labeling.

    LSTM model using RR time series as input is studied in [1]_ for
    atrial fibrillation detection. It is further improved in [2]_ via
    incorporating attention mechanism and conditional random fields.

    Parameters
    ----------
    classes : List[str]
        List of the names of the classes.
    config : dict
        Other hyper-parameters, including kernel sizes, etc.
        Refer to corresponding config file for details.

    References
    ----------
    .. [1] Faust, Oliver, et al. "Automated detection of atrial fibrillation
           using long short-term memory network with RR interval signals."
           Computers in biology and medicine 102 (2018): 327-335.
    .. [2] Wen, Hao, et al. "A Scalable Hybrid Model for Atrial Fibrillation Detection."
           Journal of Mechanics in Medicine and Biology 21.05 (2021): 2140021.

    """

    __name__ = "RR_LSTM"

    def __init__(
        self, classes: Sequence[str], config: Optional[CFG] = None, **kwargs: Any
    ) -> None:
        super().__init__()
        self.classes = list(classes)
        self.n_classes = len(classes)
        self.config = deepcopy(RR_LSTM_CONFIG)
        if not config:
            warnings.warn(
                "No config is provided, using default config.", RuntimeWarning
            )
        self.config.update(deepcopy(config) or {})
        self.config.batch_first = self.config.get("batch_first", False)

        if self.config.batch_first:
            self.in_rearrange = Rearrange(
                "batch_size n_channels seq_len -> seq_len batch_size n_channels"
            )
        else:
            self.in_rearrange = nn.Identity()

        self.lstm = StackedLSTM(
            input_size=1,
            hidden_sizes=self.config.lstm.hidden_sizes,
            bias=self.config.lstm.bias,
            dropouts=self.config.lstm.dropouts,
            bidirectional=self.config.lstm.bidirectional,
            return_sequences=self.config.lstm.retseq,
        )

        attn_input_size = self.lstm.compute_output_shape(None, None)[-1]

        if not self.config.lstm.retseq:
            if self.config.attn.name.lower() != "none":
                warnings.warn(
                    "Attention is not supported when lstm is not returning sequences",
                    RuntimeWarning,
                )
                self.config.attn.name = "none"
            self.attn_in_rearrange = nn.Identity()
            self.attn = nn.Identity()
            self.attn_out_rearrange = nn.Identity()
            clf_input_size = attn_input_size
        elif self.config.attn.name.lower() == "none":
            self.attn_in_rearrange = Rearrange(
                "seq_len batch_size n_channels -> batch_size n_channels seq_len"
            )
            self.attn = nn.Identity()
            self.attn_out_rearrange = nn.Identity()
            clf_input_size = attn_input_size
        elif self.config.attn.name.lower() == "nl":  # non_local
            self.attn_in_rearrange = Rearrange(
                "seq_len batch_size n_channels -> batch_size n_channels seq_len"
            )
            self.attn = NonLocalBlock(
                in_channels=attn_input_size,
                filter_lengths=self.config.attn.nl.filter_lengths,
                subsample_length=self.config.attn.nl.subsample_length,
                batch_norm=self.config.attn.nl.batch_norm,
            )
            self.attn_out_rearrange = nn.Identity()
            clf_input_size = self.attn.compute_output_shape(None, None)[1]
        elif self.config.attn.name.lower() == "se":  # squeeze_exitation
            self.attn_in_rearrange = Rearrange(
                "seq_len batch_size n_channels -> batch_size n_channels seq_len"
            )
            self.attn = SEBlock(
                in_channels=attn_input_size,
                reduction=self.config.attn.se.reduction,
                activation=self.config.attn.se.activation,
                kw_activation=self.config.attn.se.kw_activation,
                bias=self.config.attn.se.bias,
            )
            self.attn_out_rearrange = nn.Identity()
            clf_input_size = self.attn.compute_output_shape(None, None)[1]
        elif self.config.attn.name.lower() == "gc":  # global_context
            self.attn_in_rearrange = Rearrange(
                "seq_len batch_size n_channels -> batch_size n_channels seq_len"
            )
            self.attn = GlobalContextBlock(
                in_channels=attn_input_size,
                ratio=self.config.attn.gc.ratio,
                reduction=self.config.attn.gc.reduction,
                pooling_type=self.config.attn.gc.pooling_type,
                fusion_types=self.config.attn.gc.fusion_types,
            )
            self.attn_out_rearrange = nn.Identity()
            clf_input_size = self.attn.compute_output_shape(None, None)[1]
        elif self.config.attn.name.lower() == "sa":  # self_attention
            # NOTE: this branch NOT tested
            self.attn_in_rearrange = nn.Identity()
            self.attn = SelfAttention(
                embed_dim=attn_input_size,
                num_heads=self.config.attn.sa.get(
                    "num_heads", self.config.attn.sa.get("head_num")
                ),
                dropout=self.config.attn.sa.dropout,
                bias=self.config.attn.sa.bias,
            )
            self.attn_out_rearrange = Rearrange(
                "seq_len batch_size channels -> batch_size channels seq_len"
            )
            clf_input_size = self.attn.compute_output_shape(None, None)[-1]
        else:
            raise NotImplementedError(
                f"Attn module \042{self.config.attn.name}\042 not implemented yet."
            )

        if not self.config.lstm.retseq:
            if self.config.global_pool.lower() != "none":
                warnings.warn(
                    f"Global pooling \042{self.config.global_pool}\042 is ignored in non-sequence mode",
                    RuntimeWarning,
                )
                self.config.global_pool = "none"
            self.pool = nn.Identity()
            self.pool_rearrange = nn.Identity()
            if self.config.clf.name.lower() == "crf":
                warnings.warn(
                    "CRF layer is not supported in non-sequence mode, using linear instead.",
                    RuntimeWarning,
                )
                self.config.clf.name = "linear"
                assert (
                    "linear" in self.config.clf
                ), "Linear layer not defined in `config`."
            self.clf = MLP(
                in_channels=clf_input_size,
                out_channels=self.config.clf.linear.out_channels + [self.n_classes],
                activation=self.config.clf.linear.activation,
                bias=self.config.clf.linear.bias,
                dropouts=self.config.clf.linear.dropouts,
                skip_last_activation=True,
            )
        elif self.config.clf.name.lower() == "linear":
            if self.config.global_pool.lower() == "max":
                self.pool = nn.AdaptiveMaxPool1d(
                    (self.config.global_pool_size,), return_indices=False
                )
                self.pool_rearrange = Rearrange(
                    "batch_size channels pool_size -> batch_size (channels pool_size)"
                )
                clf_input_size *= self.config.global_pool_size
            elif self.config.global_pool.lower() == "avg":
                self.pool = nn.AdaptiveAvgPool1d((self.config.global_pool_size,))
                self.pool_rearrange = Rearrange(
                    "batch_size channels pool_size -> batch_size (channels pool_size)"
                )
                clf_input_size *= self.config.global_pool_size
            elif self.config.global_pool.lower() == "none":
                self.pool = nn.Identity()
                self.pool_rearrange = Rearrange(
                    "batch_size n_channels seq_len -> batch_size seq_len n_channels"
                )
            else:
                raise NotImplementedError(
                    f"Pooling type \042{self.config.global_pool}\042 not supported"
                )
            self.clf = MLP(
                in_channels=clf_input_size,
                out_channels=self.config.clf.linear.out_channels + [self.n_classes],
                activation=self.config.clf.linear.activation,
                bias=self.config.clf.linear.bias,
                dropouts=self.config.clf.linear.dropouts,
                skip_last_activation=True,
            )
        elif self.config.clf.name.lower() == "crf":
            if self.config.global_pool.lower() != "none":
                warnings.warn(
                    f"Global pooling \042{self.config.global_pool}\042 is ignored for CRF prediction head.",
                    RuntimeWarning,
                )
                self.config.global_pool = "none"
            self.pool = nn.Identity()
            self.pool_rearrange = Rearrange(
                "batch_size n_channels seq_len -> batch_size seq_len n_channels"
            )
            self.clf = ExtendedCRF(
                in_channels=clf_input_size,
                num_tags=self.n_classes,
                bias=self.config.clf.crf.proj_bias,
            )

        # for inference, except for crf
        self.softmax = nn.Softmax(dim=-1)
        self.sigmoid = nn.Sigmoid()

    def forward(self, input: Tensor) -> Tensor:
        """Forward pass of the model.

        Parameters
        ----------
        input : torch.Tensor
            Input RR series tensor
            of shape ``(seq_len, batch_size, n_channels)``,
            or ``(batch_size, n_channels, seq_len)``
            if `config.batch_first` is True.

        Returns
        -------
        torch.Tensor
            Output tensor,
            of shape ``(batch_size, seq_len, n_classes)``
            or ``(batch_size, n_classes)``.

        """
        x = self.in_rearrange(input)
        x = self.lstm(x)

        x = self.attn_in_rearrange(x)
        x = self.attn(x)
        x = self.attn_out_rearrange(x)

        x = self.pool(x)
        x = self.pool_rearrange(x)

        output = self.clf(x)

        return output

    @torch.no_grad()
    def inference(self, input: Tensor, bin_pred_thr: float = 0.5) -> BaseOutput:
        """Inference method for the model."""
        raise NotImplementedError("implement a task specific inference method")

    def compute_output_shape(
        self, seq_len: Optional[int] = None, batch_size: Optional[int] = None
    ) -> Sequence[Union[int, None]]:
        """Compute the output shape of the model.

        Parameters
        ----------
        seq_len : int, optional
            Length of the input series tensor.
        batch_size : int, optional
            Batch size of the input series tensor.

        Returns
        -------
        output_shape : sequence
            Output shape of the model.

        """
        if (not self.config.lstm.retseq) or (self.config.global_pool.lower() != "none"):
            output_shape = (batch_size, self.n_classes)
        else:
            output_shape = (batch_size, seq_len, self.n_classes)
        return output_shape

    @property
    def doi(self) -> List[str]:
        return list(
            set(
                self.config.get("doi", [])
                + [
                    "10.1162/neco.1997.9.8.1735",
                    "10.1016/j.compbiomed.2018.07.001",
                    "10.1142/s0219519421400212",
                ]
            )
        )

    @classmethod
    def from_v1(cls, v1_ckpt: str, device: Optional[torch.device] = None) -> "RR_LSTM":
        """Restore an instance of the model from a v1 checkpoint.

        Parameters
        ----------
        v1_ckpt : str
            Path to the v1 checkpoint file.

        Returns
        -------
        model : RR_LSTM
            The model instance restored from the v1 checkpoint.

        """
        v1_model, _ = RR_LSTM_v1.from_checkpoint(v1_ckpt, device=device)
        model = cls(classes=v1_model.classes, config=v1_model.config)
        model = model.to(v1_model.device)
        model.lstm.load_state_dict(v1_model.lstm.state_dict())
        if model.attn.__class__.__name__ != "Identity":
            model.attn.load_state_dict(v1_model.attn.state_dict())
        model.clf.load_state_dict(v1_model.clf.state_dict())
        del v1_model
        return model


class RR_LSTM_v1(nn.Module, CkptMixin, SizeMixin, CitationMixin):
    """LSTM model for RR time series classification or sequence labeling.

    LSTM model using RR time series as input is studied in [1]_ for
    atrial fibrillation detection. It is further improved in [2]_ via
    incorporating attention mechanism and conditional random fields.

    Parameters
    ----------
    classes : List[str]
        List of the names of the classes.
    config : dict
        Other hyper-parameters, including kernel sizes, etc.
        Refer to corresponding config file for details.

    References
    ----------
    .. [1] Faust, Oliver, et al. "Automated detection of atrial fibrillation
           using long short-term memory network with RR interval signals."
           Computers in biology and medicine 102 (2018): 327-335.
    .. [2] Wen, Hao, et al. "A Scalable Hybrid Model for Atrial Fibrillation Detection."
           Journal of Mechanics in Medicine and Biology 21.05 (2021): 2140021.

    """

    __name__ = "RR_LSTM_v1"

    def __init__(
        self, classes: Sequence[str], config: Optional[CFG] = None, **kwargs: Any
    ) -> None:
        """
        Parameters
        ----------
        classes: list,
            list of the classes for classification
        config: dict, optional,
            other hyper-parameters, including kernel sizes, etc.
            ref. the corresponding config file

        """
        super().__init__()
        self.classes = list(classes)
        self.n_classes = len(classes)
        self.config = deepcopy(RR_LSTM_CONFIG)
        if not config:
            warnings.warn(
                "No config is provided, using default config.", RuntimeWarning
            )
        self.config.update(deepcopy(config) or {})

        self.lstm = StackedLSTM(
            input_size=1,
            hidden_sizes=self.config.lstm.hidden_sizes,
            bias=self.config.lstm.bias,
            dropouts=self.config.lstm.dropouts,
            bidirectional=self.config.lstm.bidirectional,
            return_sequences=self.config.lstm.retseq,
        )

        attn_input_size = self.lstm.compute_output_shape(None, None)[-1]

        if not self.config.lstm.retseq:
            if self.config.attn.name.lower() != "none":
                warnings.warn(
                    "Attention is not supported when lstm is not returning sequences",
                    RuntimeWarning,
                )
                self.config.attn.name = "none"
            self.attn = None
            clf_input_size = attn_input_size
        elif self.config.attn.name.lower() == "none":
            self.attn = None
            clf_input_size = attn_input_size
        elif self.config.attn.name.lower() == "nl":  # non_local
            self.attn = NonLocalBlock(
                in_channels=attn_input_size,
                filter_lengths=self.config.attn.nl.filter_lengths,
                subsample_length=self.config.attn.nl.subsample_length,
                batch_norm=self.config.attn.nl.batch_norm,
            )
            clf_input_size = self.attn.compute_output_shape(None, None)[1]
        elif self.config.attn.name.lower() == "se":  # squeeze_exitation
            self.attn = SEBlock(
                in_channels=attn_input_size,
                reduction=self.config.attn.se.reduction,
                activation=self.config.attn.se.activation,
                kw_activation=self.config.attn.se.kw_activation,
                bias=self.config.attn.se.bias,
            )
            clf_input_size = self.attn.compute_output_shape(None, None)[1]
        elif self.config.attn.name.lower() == "gc":  # global_context
            self.attn = GlobalContextBlock(
                in_channels=attn_input_size,
                ratio=self.config.attn.gc.ratio,
                reduction=self.config.attn.gc.reduction,
                pooling_type=self.config.attn.gc.pooling_type,
                fusion_types=self.config.attn.gc.fusion_types,
            )
            clf_input_size = self.attn.compute_output_shape(None, None)[1]
        elif self.config.attn.name.lower() == "sa":  # self_attention
            # NOTE: this branch NOT tested
            self.attn = SelfAttention(
                embed_dim=attn_input_size,
                num_heads=self.config.attn.sa.get(
                    "num_heads", self.config.attn.sa.get("head_num")
                ),
                dropout=self.config.attn.sa.dropout,
                bias=self.config.attn.sa.bias,
            )
            clf_input_size = self.attn.compute_output_shape(None, None)[-1]
        else:
            raise NotImplementedError(
                f"Attn module \042{self.config.attn.name}\042 not implemented yet."
            )

        if not self.config.lstm.retseq:
            if self.config.global_pool.lower() != "none":
                warnings.warn(
                    f"Global pooling \042{self.config.global_pool}\042 is ignored in non-sequence mode",
                    RuntimeWarning,
                )
                self.config.global_pool = "none"
            self.pool = None
            if self.config.clf.name.lower() == "crf":
                warnings.warn(
                    "CRF layer is not supported in non-sequence mode, using linear instead.",
                    RuntimeWarning,
                )
                self.config.clf.name = "linear"
                assert (
                    "linear" in self.config.clf
                ), "Linear layer not defined in `config`."
            self.clf = MLP(
                in_channels=clf_input_size,
                out_channels=self.config.clf.linear.out_channels + [self.n_classes],
                activation=self.config.clf.linear.activation,
                bias=self.config.clf.linear.bias,
                dropouts=self.config.clf.linear.dropouts,
                skip_last_activation=True,
            )
        elif self.config.clf.name.lower() == "linear":
            if self.config.global_pool.lower() == "max":
                self.pool = nn.AdaptiveMaxPool1d(
                    (self.config.global_pool_size,), return_indices=False
                )
                clf_input_size *= self.config.global_pool_size
            elif self.config.global_pool.lower() == "avg":
                self.pool = nn.AdaptiveAvgPool1d((self.config.global_pool_size,))
                clf_input_size *= self.config.global_pool_size
            elif self.config.global_pool.lower() == "none":
                self.pool = None
            else:
                raise NotImplementedError(
                    f"Pooling type \042{self.config.global_pool}\042 not supported"
                )
            self.clf = MLP(
                in_channels=clf_input_size,
                out_channels=self.config.clf.linear.out_channels + [self.n_classes],
                activation=self.config.clf.linear.activation,
                bias=self.config.clf.linear.bias,
                dropouts=self.config.clf.linear.dropouts,
                skip_last_activation=True,
            )
        elif self.config.clf.name.lower() == "crf":
            if self.config.global_pool.lower() != "none":
                warnings.warn(
                    f"Global pooling \042{self.config.global_pool}\042 is ignored for CRF prediction head.",
                    RuntimeWarning,
                )
                self.config.global_pool = "none"
            self.pool = None
            self.clf = ExtendedCRF(
                in_channels=clf_input_size,
                num_tags=self.n_classes,
                bias=self.config.clf.crf.proj_bias,
            )

        # for inference, except for crf
        self.softmax = nn.Softmax(dim=-1)
        self.sigmoid = nn.Sigmoid()

    def forward(self, input: Tensor) -> Tensor:
        """Forward pass of the model.

        Parameters
        ----------
        input : torch.Tensor
            Input RR series tensor
            of shape ``(seq_len, batch_size, n_channels)``,
            or ``(batch_size, n_channels, seq_len)``
            if `config.batch_first` is True.

        Returns
        -------
        torch.Tensor
            Output tensor,
            of shape ``(batch_size, seq_len, n_classes)``
            or ``(batch_size, n_classes)``.

        """
        # (batch_size, n_channels, seq_len) --> (seq_len, batch_size, n_channels)
        # x = input.permute(1,2,0)
        x = self.lstm(
            input
        )  # (seq_len, batch_size, n_channels) or (batch_size, n_channels)
        if self.attn:
            # (seq_len, batch_size, n_channels) --> (batch_size, n_channels, seq_len)
            x = x.permute(1, 2, 0)
            x = self.attn(x)  # (batch_size, n_channels, seq_len)
        elif x.ndim == 3:
            # (seq_len, batch_size, n_channels) --> (batch_size, n_channels, seq_len)
            x = x.permute(1, 2, 0)
        if self.pool:
            x = self.pool(x)  # (batch_size, n_channels, 1)
            # x = x.squeeze(dim=-1)  # (batch_size, n_channels)
            x = rearrange(
                x,
                "batch_size n_channels pool_size -> batch_size (n_channels pool_size)",
            )
        elif x.ndim == 3:
            x = x.permute(
                0, 2, 1
            )  # (batch_size, n_channels, seq_len) --> (batch_size, seq_len, n_channels)
        else:
            # x of shape (batch_size, n_channels),
            # in the case where config.lstm.retseq = False
            pass
        if self.config.clf.name.lower() == "linear":
            x = self.clf(
                x
            )  # (batch_size, seq_len, n_classes) or (batch_size, n_classes)
        elif self.config.clf.name.lower() == "crf":
            x = self.clf(x)  # (batch_size, seq_len, n_classes)
        output = x

        return output

    @torch.no_grad()
    def inference(self, input: Tensor, bin_pred_thr: float = 0.5) -> BaseOutput:
        """Inference method for the model."""
        raise NotImplementedError("implement a task specific inference method")

    def compute_output_shape(
        self, seq_len: Optional[int] = None, batch_size: Optional[int] = None
    ) -> Sequence[Union[int, None]]:
        """Compute the output shape of the model.

        Parameters
        ----------
        seq_len : int, optional
            Length of the input series tensor.
        batch_size : int, optional
            Batch size of the input series tensor.

        Returns
        -------
        output_shape : sequence
            Output shape of the model.

        """
        if (not self.config.lstm.retseq) or (self.config.global_pool.lower() != "none"):
            output_shape = (batch_size, self.n_classes)
        else:
            output_shape = (batch_size, seq_len, self.n_classes)
        return output_shape

    @property
    def doi(self) -> List[str]:
        return list(
            set(
                self.config.get("doi", [])
                + [
                    "10.1162/neco.1997.9.8.1735",
                    "10.1016/j.compbiomed.2018.07.001",
                    "10.1142/s0219519421400212",
                ]
            )
        )



"""
utilities for downloading and extracting compressed data files

For most PhysioNet databases, the WFDB package already has a method `dl_database`
for downloading the data files.

"""

import os
import re
import shutil
import tarfile
import tempfile
import zipfile
import warnings
import urllib
from pathlib import Path
from typing import Optional, Union, Iterable

import requests
from tqdm.auto import tqdm


__all__ = [
    "http_get",
]


PHYSIONET_DB_VERSION_PATTERN = "\\d+\\.\\d+\\.\\d+"


def http_get(
    url: str,
    dst_dir: Union[str, Path],
    proxies: Optional[dict] = None,
    extract: bool = True,
    filename: Optional[str] = None,
) -> None:
    """Get contents of a URL and save to a file.

    This function is a modified version of the `download_file` function in
    `transformers.file_utils` [1]_.

    Parameters
    ----------
    url : str
        URL to download file from.
    dst_dir : str or pathlib.Path
        Directory to save the file.
    proxies : dict, optional
        Dictionary of proxy settings.
    extract : bool, default True
        Whether to extract the downloaded file.
    filename : str, optional
        Name of the file to save.
        If None, the filename will be the same as the URL.

    Returns
    -------
    None

    References
    ----------
    .. [1] https://github.com/huggingface/transformers/blob/master/src/transformers/file_utils.py

    """
    if filename is not None:
        assert not (Path(dst_dir) / filename).exists(), "file already exists"
    print(f"Downloading {url}.")
    if not is_compressed_file(url) and extract:
        if filename is not None:
            if not is_compressed_file(filename):
                warnings.warn(
                    "filename is given, and it is not a `zip` file or a compressed `tar` file. "
                    "Automatic decompression is turned off.",
                    RuntimeWarning,
                )
                extract = False
            else:
                pass
        else:
            warnings.warn(
                "URL must be pointing to a `zip` file or a compressed `tar` file. "
                "Automatic decompression is turned off. "
                "The user is responsible for decompressing the file manually.",
                RuntimeWarning,
            )
            extract = False
    # for example "https://www.dropbox.com/s/xxx/test%3F.zip??dl=1"
    # produces pure_url = "https://www.dropbox.com/s/xxx/test?.zip"
    pure_url = urllib.parse.unquote(url.split("?")[0])
    parent_dir = Path(dst_dir).parent
    df_suffix = _suffix(pure_url) if filename is None else _suffix(filename)
    downloaded_file = tempfile.NamedTemporaryFile(
        dir=parent_dir,
        suffix=df_suffix,
        delete=False,
    )
    req = requests.get(url, stream=True, proxies=proxies)
    content_length = req.headers.get("Content-Length")
    total = int(content_length) if content_length is not None else None
    if req.status_code == 403 or req.status_code == 404:
        raise Exception(f"Could not reach {url}.")
    progress = tqdm(
        unit="B", unit_scale=True, total=total, dynamic_ncols=True, mininterval=1.0
    )
    for chunk in req.iter_content(chunk_size=1024):
        if chunk:  # filter out keep-alive new chunks
            progress.update(len(chunk))
            downloaded_file.write(chunk)
    progress.close()
    downloaded_file.close()
    if extract:
        if ".zip" in df_suffix:
            _unzip_file(str(downloaded_file.name), str(dst_dir))
        elif ".tar" in df_suffix:  # tar files
            _untar_file(str(downloaded_file.name), str(dst_dir))
        else:
            os.remove(downloaded_file.name)
            raise Exception(f"Unknown file type {df_suffix}")
        # avoid the case the compressed file is a folder with the same name
        # DO NOT use _stem(Path(pure_url))
        if filename is None:
            _folder = Path(url).name.replace(_suffix(url), "")
        else:
            _folder = _stem(Path(filename))
        if _folder in os.listdir(dst_dir):
            tmp_folder = str(dst_dir).rstrip(os.sep) + "_tmp"
            os.rename(dst_dir, tmp_folder)
            os.rename(Path(tmp_folder) / _folder, dst_dir)
            shutil.rmtree(tmp_folder)
    else:
        Path(dst_dir).mkdir(parents=True, exist_ok=True)
        if filename is None:
            shutil.copyfile(downloaded_file.name, Path(dst_dir) / Path(pure_url).name)
        else:  # filename is not None
            shutil.copyfile(downloaded_file.name, Path(dst_dir) / filename)
    os.remove(downloaded_file.name)


def _stem(path: Union[str, Path]) -> str:
    """Get filename without extension,
    especially for .tar.xx files.

    Parameters
    ----------
    path : str or pathlib.Path
        Path to the file

    Returns
    -------
    str
        Filename without extension.

    """
    ret = Path(path).stem
    if Path(ret).suffix in [".tar", ".gz", ".bz2", ".xz", ".zip", ".7z"]:
        return _stem(ret)
    return ret


def _suffix(
    path: Union[str, Path], ignore_pattern: str = PHYSIONET_DB_VERSION_PATTERN
) -> str:
    """Get file extension, including all suffixes.

    Parameters
    ----------
    path : str or pathlib.Path
        Path to the file.
    ignore_pattern : str, optional
        Pattern to ignore in the filename,
        by default `PHYSIONET_DB_VERSION_PATTERN`.

    Returns
    -------
    str
        The full file extension.

    """
    return "".join(Path(re.sub(ignore_pattern, "", str(path))).suffixes)


def is_compressed_file(path: Union[str, Path]) -> bool:
    """Check if the path points to a valid compressed file.

    Parameters
    ----------
    path : str or pathlib.Path
        Path to the file

    Returns
    -------
    bool
        True if the file is a valid compressed file,
        False otherwise.

    """
    compressed_file_pattern = "(\\.zip)|(\\.tar)"
    return re.search(compressed_file_pattern, _suffix(path)) is not None


def _unzip_file(path_to_zip_file: Union[str, Path], dst_dir: Union[str, Path]) -> None:
    """Unzips a .zip file to folder path.

    Parameters
    ----------
    path_to_zip_file : str or pathlib.Path
        Path to the .zip file.
    dst_dir : str or pathlib.Path
        Path to the destination folder.

    Returns
    -------
    None

    """
    print(f"Extracting file {path_to_zip_file} to {dst_dir}.")
    with zipfile.ZipFile(str(path_to_zip_file)) as zip_ref:
        zip_ref.extractall(str(dst_dir))


def _untar_file(path_to_tar_file: Union[str, Path], dst_dir: Union[str, Path]) -> None:
    """Decompress a .tar.xx file to folder path.

    Parameters
    ----------
    path_to_tar_file : str or pathlib.Path
        Path to the .tar.xx file.
    dst_dir : str or pathlib.Path
        Path to the destination folder.

    Returns
    -------
    None

    """
    print(f"Extracting file {path_to_tar_file} to {dst_dir}.")
    mode = (
        Path(path_to_tar_file).suffix.replace(".", "r:").replace("tar", "").strip(":")
    )
    with tarfile.open(str(path_to_tar_file), mode) as tar_ref:
        # tar_ref.extractall(str(dst_dir))
        # CVE-2007-4559 (related to  CVE-2001-1267):
        # directory traversal vulnerability in `extract` and `extractall` in `tarfile` module
        _safe_tar_extract(tar_ref, str(dst_dir))


def _is_within_directory(directory: Union[str, Path], target: Union[str, Path]) -> bool:
    """Check if the target is within the directory.

    Parameters
    ----------
    directory : str or pathlib.Path
        Directory to check.
    target : str or pathlib.Path
        Path to the target.

    Returns
    -------
    bool
        True if the target is within the directory,
        False otherwise.

    """
    abs_directory = os.path.abspath(directory)
    abs_target = os.path.abspath(target)

    prefix = os.path.commonprefix([abs_directory, abs_target])

    return prefix == abs_directory


def _safe_tar_extract(
    tar: tarfile.TarFile,
    dst_dir: Union[str, Path],
    members: Optional[Iterable[tarfile.TarInfo]] = None,
    *,
    numeric_owner: bool = False,
) -> None:
    """Extract members from a tarfile **safely**
    to a destination directory.

    This function prevents path traversal attacks, by checking that the
    extracted files are within the destination directory.

    Parameters
    ----------
    tar : tarfile.TarFile
        The tarfile to extract from.
    dst_dir : str or pathlib.Path
        The destination directory
    members : Iterable[tarfile.TarInfo], optional
        The members to extract
        If is None, extract all members,
        otherwise, must be a subset of the list
        returned by :func:`tar.getmembers`.
    numeric_owner : bool, default False
        If True, only the numbers for user/group names are used
        and not the names. For more information,
        see :func:`tarfile.TarFile.extractall`.

    """
    for member in members or tar.getmembers():
        member_path = os.path.join(dst_dir, member.name)
        if not _is_within_directory(dst_dir, member_path):
            raise Exception("Attempted Path Traversal in Tar File")

    tar.extractall(dst_dir, members, numeric_owner=numeric_owner)



"""
"""

from ....cfg import CFG, DEFAULTS

__all__ = [
    "CPSC2019TrainCfg",
]


_NAME = "cinc2021"


CPSC2019TrainCfg = CFG()

CPSC2019TrainCfg.fs = 500  # Hz, CPSC2019 data fs
CPSC2019TrainCfg.classes = [
    "N",
]
CPSC2019TrainCfg.n_leads = 1

CPSC2019TrainCfg.db_dir = None
CPSC2019TrainCfg.log_dir = DEFAULTS.log_dir / _NAME
CPSC2019TrainCfg.checkpoints = DEFAULTS.checkpoints / _NAME
CPSC2019TrainCfg.model_dir = DEFAULTS.model_dir / _NAME
CPSC2019TrainCfg.log_dir.mkdir(parents=True, exist_ok=True)
CPSC2019TrainCfg.checkpoints.mkdir(parents=True, exist_ok=True)
CPSC2019TrainCfg.model_dir.mkdir(parents=True, exist_ok=True)

CPSC2019TrainCfg.bias_thr = (
    0.075 * CPSC2019TrainCfg.fs
)  # keep the same with `THR` in `cpsc2019_score.py`
# detected rpeaks that are within `skip_dist` from two ends of the signal will be ignored,
# as in the official entry function
CPSC2019TrainCfg.skip_dist = 0.5 * CPSC2019TrainCfg.fs
CPSC2019TrainCfg.torch_dtype = DEFAULTS.DTYPE.TORCH
CPSC2019TrainCfg.np_dtype = DEFAULTS.DTYPE.NP

CPSC2019TrainCfg.final_model_name = None
CPSC2019TrainCfg.keep_checkpoint_max = 20
CPSC2019TrainCfg.train_ratio = 0.8

CPSC2019TrainCfg.input_len = int(CPSC2019TrainCfg.fs * 10)  # 10 s

# configs of signal preprocessing
CPSC2019TrainCfg.normalize = False
# frequency band of the filter to apply, should be chosen very carefully
# CPSC2019TrainCfg.bandpass = False
CPSC2019TrainCfg.bandpass = CFG(
    lowcut=0.5,
    highcut=45,
)

# configs of data aumentation
# NOTE: compared to data augmentation of CPSC2020,
# `stretch_compress` and `label_smoothing` are not used in CPSC2019
CPSC2019TrainCfg.label_smooth = False
CPSC2019TrainCfg.random_masking = False
CPSC2019TrainCfg.stretch_compress = False  # stretch or compress in time axis
CPSC2019TrainCfg.mixup = False
CPSC2019TrainCfg.random_flip = CFG(
    prob=0.5,
)

# configs of training epochs, batch, etc.
CPSC2019TrainCfg.n_epochs = 150
CPSC2019TrainCfg.batch_size = 32

# configs of optimizers and lr_schedulers
CPSC2019TrainCfg.optimizer = "adamw_amsgrad"  # "sgd", "adam", "adamw"
CPSC2019TrainCfg.momentum = 0.949  # default values for corresponding PyTorch optimizers
CPSC2019TrainCfg.betas = (
    0.9,
    0.999,
)  # default values for corresponding PyTorch optimizers
CPSC2019TrainCfg.decay = 1e-2  # default values for corresponding PyTorch optimizers

CPSC2019TrainCfg.learning_rate = 1e-3  # 1e-4
CPSC2019TrainCfg.lr = CPSC2019TrainCfg.learning_rate

CPSC2019TrainCfg.lr_scheduler = (
    "one_cycle"  # "one_cycle", "plateau", "burn_in", "step", None
)
CPSC2019TrainCfg.lr_step_size = 50
CPSC2019TrainCfg.lr_gamma = 0.1
CPSC2019TrainCfg.max_lr = 2e-3  # for "one_cycle" scheduler, to adjust via expriments

CPSC2019TrainCfg.burn_in = 400
CPSC2019TrainCfg.steps = [5000, 10000]

CPSC2019TrainCfg.early_stopping = CFG()  # early stopping according to challenge metric
CPSC2019TrainCfg.early_stopping.min_delta = 0.0001  # should be non-negative
CPSC2019TrainCfg.early_stopping.patience = 15

# configs of loss function
CPSC2019TrainCfg.loss = "BCEWithLogitsLoss"
# CPSC2019TrainCfg.loss = "BCEWithLogitsWithClassWeightLoss"
CPSC2019TrainCfg.flooding_level = 0.0  # flooding performed if positive

CPSC2019TrainCfg.log_step = 2

# model selection
CPSC2019TrainCfg.model_name = "seq_lab_crnn"  # "seq_lab_cnn", "unet", "subtract_unet"
CPSC2019TrainCfg.cnn_name = "multi_scopic"
CPSC2019TrainCfg.rnn_name = "lstm"
CPSC2019TrainCfg.attn_name = "se"

CPSC2019TrainCfg.reduction = 2**3  # TODO: automatic adjust via model config
CPSC2019TrainCfg.recover_length = True

CPSC2019TrainCfg.monitor = "qrs_score"



"""
vanilla resnet, and various variants

Reference
---------
1. He, K., Zhang, X., Ren, S., & Sun, J. (2016). Deep residual learning for image recognition. In Proceedings of the IEEE conference on computer vision and pattern recognition (pp. 770-778).
2. He, T., Zhang, Z., Zhang, H., Zhang, Z., Xie, J., & Li, M. (2019). Bag of tricks for image classification with convolutional neural networks. In Proceedings of the IEEE/CVF Conference on Computer Vision and Pattern Recognition (pp. 558-567).
3. Hannun, A. Y., Rajpurkar, P., Haghpanahi, M., Tison, G. H., Bourn, C., Turakhia, M. P., & Ng, A. Y. (2019). Cardiologist-level arrhythmia detection and classification in ambulatory electrocardiograms using a deep neural network. Nature medicine, 25(1), 65-69.
4. Ribeiro, A. H., Ribeiro, M. H., Paixão, G. M., Oliveira, D. M., Gomes, P. R., Canazart, J. A., ... & Ribeiro, A. L. P. (2020). Automatic diagnosis of the 12-lead ECG using a deep neural network. Nature communications, 11(1), 1-9.
5. Ridnik, T., Lawen, H., Noy, A., Ben Baruch, E., Sharir, G., & Friedman, I. (2021). Tresnet: High performance gpu-dedicated architecture. In Proceedings of the IEEE/CVF Winter Conference on Applications of Computer Vision (pp. 1400-1409).
"""

from copy import deepcopy

from ...cfg import CFG
from ..attn import global_context, non_local, squeeze_excitation

__all__ = [
    # building blocks
    "resnet_block_basic",
    "resnet_bottle_neck",
    "resnet_bottle_neck_B",
    "resnet_bottle_neck_D",
    "resnet_block_basic_se",
    "resnet_bottle_neck_se",
    "resnet_block_basic_nl",
    "resnet_bottle_neck_nl",
    "resnet_block_basic_gc",
    "resnet_bottle_neck_gc",
    # smaller resnets
    "resnetN",
    "resnetNB",
    "resnetNS",
    "resnetNBS",
    # vanilla resnet
    "resnet_vanilla_18",
    "resnet_vanilla_34",
    "resnet_vanilla_50",
    "resnet_vanilla_101",
    "resnet_vanilla_152",
    "resnext_vanilla_50_32x4d",
    "resnext_vanilla_101_32x8d",
    "resnet_vanilla_wide_50_2",
    "resnet_vanilla_wide_101_2",
    # cpsc2018 resnet
    "resnet_cpsc2018",
    "resnet_cpsc2018_leadwise",
    # stanford resnet
    "resnet_block_stanford",
    "resnet_stanford",
    # ResNet Nature Communications
    "resnet_nature_comm",
    "resnet_nature_comm_se",
    "resnet_nature_comm_nl",
    "resnet_nature_comm_gc",
    "resnet_nature_comm_bottle_neck",
    "resnet_nature_comm_bottle_neck_se",
    "resnet_nature_comm_bottle_neck_gc",
    "resnet_nature_comm_bottle_neck_nl",
    # TresNet
    "tresnetF",
    "tresnetP",
    "tresnetN",
    "tresnetS",
    "tresnetM",
    "tresnetL",
    "tresnetXL",
    "tresnetM_V2",
]


# building blocks

resnet_block_basic = CFG()
resnet_block_basic.increase_channels_method = "conv"  # or "zero_padding"
resnet_block_basic.subsample_mode = (
    "conv"  # or "max", "avg", "nearest", "linear", "bilinear"
)
resnet_block_basic.kernel_initializer = "he_normal"
resnet_block_basic.kw_initializer = {}
resnet_block_basic.activation = "relu"  # "mish", "swish"
resnet_block_basic.kw_activation = {"inplace": True}
resnet_block_basic.bias = False

resnet_bottle_neck = CFG()
resnet_bottle_neck.expansion = 4
resnet_bottle_neck.increase_channels_method = "conv"  # or "zero_padding"
resnet_bottle_neck.subsample_mode = (
    "conv"  # or "max", "avg", "nearest", "linear", "bilinear"
)
resnet_bottle_neck.subsample_at = (
    0  # or 0  # 0 is for the original paper, 1 for ResNet-B
)
resnet_bottle_neck.kernel_initializer = "he_normal"
resnet_bottle_neck.kw_initializer = {}
resnet_bottle_neck.activation = "relu"  # "mish", "swish"
resnet_bottle_neck.kw_activation = {"inplace": True}
resnet_bottle_neck.bias = False

resnet_bottle_neck_B = deepcopy(resnet_bottle_neck)
resnet_bottle_neck_B.subsample_at = 1

resnet_bottle_neck_D = deepcopy(resnet_bottle_neck_B)
resnet_bottle_neck_D.subsample_mode = "avg"

# one can add more variants by changing the attention modules
resnet_block_basic_se = deepcopy(resnet_block_basic)
resnet_block_basic_se.attn = deepcopy(squeeze_excitation)
resnet_block_basic_se.attn.name = "se"
resnet_block_basic_se.attn.pos = -1

resnet_bottle_neck_se = deepcopy(resnet_bottle_neck_B)
resnet_bottle_neck_se.attn = deepcopy(squeeze_excitation)
resnet_bottle_neck_se.attn.name = "se"
resnet_bottle_neck_se.attn.pos = -1

resnet_block_basic_nl = deepcopy(resnet_block_basic)
resnet_block_basic_nl.attn = deepcopy(non_local)
resnet_block_basic_nl.attn.name = "nl"
resnet_block_basic_nl.attn.pos = -1

resnet_bottle_neck_nl = deepcopy(resnet_bottle_neck_B)
resnet_bottle_neck_nl.attn = deepcopy(non_local)
resnet_bottle_neck_nl.attn.name = "nl"
resnet_bottle_neck_nl.attn.pos = -1

resnet_block_basic_gc = deepcopy(resnet_block_basic)
resnet_block_basic_gc.attn = deepcopy(global_context)
resnet_block_basic_gc.attn.name = "gc"
resnet_block_basic_gc.attn.pos = -1

resnet_bottle_neck_gc = deepcopy(resnet_bottle_neck_B)
resnet_bottle_neck_gc.attn = deepcopy(global_context)
resnet_bottle_neck_gc.attn.name = "gc"
resnet_bottle_neck_gc.attn.pos = -1


resnet_stem = CFG()
resnet_stem.num_filters = 64
resnet_stem.filter_lengths = 25
resnet_stem.conv_stride = 2
resnet_stem.pool_size = 3
resnet_stem.pool_stride = 2

resnet_stem_C = deepcopy(resnet_stem)
resnet_stem_C.num_filters = [
    32,
    32,
    64,
]
resnet_stem_C.filter_lengths = [
    15,
    15,
    15,
]


# vanilla ResNets
resnet_vanilla_common = CFG()
resnet_vanilla_common.fs = 500
resnet_vanilla_common.subsample_lengths = [
    1,
    2,
    2,
    2,
]
resnet_vanilla_common.filter_lengths = 15
resnet_vanilla_common.groups = 1
resnet_vanilla_common.increase_channels_method = "conv"
resnet_vanilla_common.kernel_initializer = "he_normal"
resnet_vanilla_common.kw_initializer = {}
resnet_vanilla_common.activation = "relu"  # "mish", "swish"
resnet_vanilla_common.kw_activation = {"inplace": True}
resnet_vanilla_common.bias = False
resnet_vanilla_common.stem = deepcopy(resnet_stem)


resnet_vanilla_18 = CFG()
resnet_vanilla_18.num_blocks = [
    2,
    2,
    2,
    2,
]
resnet_vanilla_18.update(deepcopy(resnet_vanilla_common))
resnet_vanilla_18.building_block = "basic"
resnet_vanilla_18.block = deepcopy(resnet_block_basic)
resnet_vanilla_18.block.kernel_initializer = resnet_vanilla_18.kernel_initializer
resnet_vanilla_18.block.kw_initializer = resnet_vanilla_18.kw_initializer
resnet_vanilla_18.block.activation = resnet_vanilla_18.activation
resnet_vanilla_18.block.kw_activation = resnet_vanilla_18.kw_activation
resnet_vanilla_18.block.bias = resnet_vanilla_18.bias

resnet_vanilla_34 = deepcopy(resnet_vanilla_18)
resnet_vanilla_34.num_blocks = [
    3,
    4,
    6,
    3,
]

resnet_vanilla_50 = CFG()  # uses bottleneck
resnet_vanilla_50.num_blocks = [
    3,
    4,
    6,
    3,
]
resnet_vanilla_50.update(deepcopy(resnet_vanilla_common))
resnet_vanilla_50.base_groups = 1
resnet_vanilla_50.base_width = 64
resnet_vanilla_50.building_block = "bottleneck"
resnet_vanilla_50.block = deepcopy(resnet_bottle_neck_B)
resnet_vanilla_50.block.kernel_initializer = resnet_vanilla_50.kernel_initializer
resnet_vanilla_50.block.kw_initializer = resnet_vanilla_50.kw_initializer
resnet_vanilla_50.block.activation = resnet_vanilla_50.activation
resnet_vanilla_50.block.kw_activation = resnet_vanilla_50.kw_activation
resnet_vanilla_50.block.bias = resnet_vanilla_50.bias

resnet_vanilla_101 = deepcopy(resnet_vanilla_50)  # uses bottleneck
resnet_vanilla_101.num_blocks = [
    3,
    4,
    23,
    3,
]
resnet_vanilla_101.filter_lengths = 11

resnet_vanilla_152 = deepcopy(resnet_vanilla_101)  # uses bottleneck
resnet_vanilla_152.num_blocks = [
    3,
    8,
    36,
    3,
]

resnext_vanilla_50_32x4d = deepcopy(resnet_vanilla_152)  # uses bottleneck
resnext_vanilla_50_32x4d.num_blocks = [
    3,
    4,
    6,
    3,
]
resnext_vanilla_50_32x4d.groups = 32
resnext_vanilla_50_32x4d.base_groups = 1
resnext_vanilla_50_32x4d.base_width = 4

resnext_vanilla_101_32x8d = deepcopy(resnext_vanilla_50_32x4d)  # uses bottleneck
resnext_vanilla_101_32x8d.num_blocks = [
    3,
    4,
    23,
    3,
]
resnext_vanilla_101_32x8d.base_width = 8

resnet_vanilla_wide_50_2 = deepcopy(resnet_vanilla_50)  # uses bottleneck
resnet_vanilla_wide_50_2.base_groups = 1
resnet_vanilla_wide_50_2.base_width = 64 * 2

resnet_vanilla_wide_101_2 = deepcopy(resnet_vanilla_101)  # uses bottleneck
resnet_vanilla_wide_101_2.base_groups = 1
resnet_vanilla_wide_101_2.base_width = 64 * 2


# smaller ResNets
resnetN = deepcopy(resnet_vanilla_18)
resnetN.num_blocks = [
    1,
    1,
    1,
    1,
]
resnetN.filter_lengths = [
    19,
    15,
    11,
    7,
]

resnetNB = deepcopy(resnetN)
resnetNB.building_block = "bottleneck"
resnetNB.block = deepcopy(resnet_bottle_neck_B)

resnetNS = deepcopy(resnetN)
resnetNS.block.conv_type = "separable"

resnetNBS = deepcopy(resnetNB)
resnetNBS.block.conv_type = "separable"


# custom ResNets
resnet_cpsc2018 = CFG()
resnet_cpsc2018.fs = 500
resnet_cpsc2018.building_block = "basic"  # "bottleneck"
resnet_cpsc2018.expansion = 1
resnet_cpsc2018.subsample_lengths = [
    1,
    2,
    2,
    2,
]
# resnet_cpsc2018.num_blocks = [
#     2, 2, 2, 2, 2,
# ]
# resnet_cpsc2018.filter_lengths = 3
# resnet_cpsc2018.num_blocks = [
#     3, 4, 6, 3,
# ]
# resnet_cpsc2018.filter_lengths = [
#     [5, 5, 13],
#     [5, 5, 5, 13],
#     [5, 5, 5, 5, 5, 13],
#     [5, 5, 25],
# ]
resnet_cpsc2018.num_blocks = [
    3,
    4,
    6,
    3,
]
resnet_cpsc2018.filter_lengths = [
    [5, 5, 25],
    [5, 5, 5, 25],
    [5, 5, 5, 5, 5, 25],
    [5, 5, 49],
]
resnet_cpsc2018.groups = 1
_base_num_filters = 12 * 4
resnet_cpsc2018.stem = deepcopy(resnet_stem)
resnet_cpsc2018.stem.num_filters = _base_num_filters
resnet_cpsc2018.stem.filter_lengths = 15  # corr. to 30 ms
resnet_cpsc2018.kernel_initializer = "he_normal"
resnet_cpsc2018.kw_initializer = {}
resnet_cpsc2018.activation = "relu"  # "mish", "swish"
resnet_cpsc2018.kw_activation = {"inplace": True}
resnet_cpsc2018.bias = False


resnet_cpsc2018_leadwise = deepcopy(resnet_cpsc2018)
resnet_cpsc2018_leadwise.groups = 12
resnet_cpsc2018_leadwise.stem.num_filters = 12 * 8


# set default building block
resnet_cpsc2018.building_block = "basic"
resnet_cpsc2018.block = deepcopy(resnet_block_basic)
resnet_cpsc2018_leadwise.building_block = "basic"
resnet_cpsc2018_leadwise.block = deepcopy(resnet_block_basic)


# ResNet Stanford
resnet_stanford = CFG()
resnet_stanford.fs = 500
resnet_stanford.groups = 1
resnet_stanford.num_blocks = [
    2,
    2,
    2,
    2,
]
resnet_stanford.subsample_lengths = 2
resnet_stanford.filter_lengths = 17
_base_num_filters = 36
resnet_stanford.stem = deepcopy(resnet_stem)
resnet_stanford.stem.num_filters = _base_num_filters * 2
resnet_stanford.stem.filter_lengths = 17
resnet_stanford.kernel_initializer = "he_normal"
resnet_stanford.kw_initializer = {}
resnet_stanford.activation = "relu"
resnet_stanford.kw_activation = {"inplace": True}
resnet_stanford.bias = False


resnet_block_stanford = CFG()
resnet_block_stanford.increase_channels_at = 4
resnet_block_stanford.increase_channels_method = "conv"  # or "zero_padding"
resnet_block_stanford.num_skip = 2
resnet_block_stanford.subsample_mode = "conv"  # "max", "avg"
resnet_block_stanford.kernel_initializer = resnet_stanford.kernel_initializer
resnet_block_stanford.kw_initializer = deepcopy(resnet_stanford.kw_initializer)
resnet_block_stanford.activation = resnet_stanford.activation
resnet_block_stanford.kw_activation = deepcopy(resnet_stanford.kw_activation)
resnet_block_stanford.dropout = 0.2
resnet_block_stanford.bias = False

resnet_stanford.building_block = "basic"
resnet_stanford.block = deepcopy(resnet_block_stanford)


# ResNet Nature Communications
# a modified version of resnet [1,1,1,1]
resnet_nature_comm = deepcopy(resnet_vanilla_18)
resnet_nature_comm.num_blocks = [
    1,
    1,
    1,
    1,
]
resnet_nature_comm.stem.filter_lengths = 17  # originally 16, we make it odd
resnet_nature_comm.filter_lengths = 17
resnet_nature_comm.stem.conv_stride = 1
resnet_nature_comm.stem.pool_stride = 1  # if < 2, init pool will not be added
resnet_nature_comm.dropouts = 0.2
resnet_nature_comm.subsample_lengths = [
    4,
    4,
    4,
    4,
]
resnet_nature_comm.num_filters = [
    2 * resnet_nature_comm.stem.num_filters,
    3 * resnet_nature_comm.stem.num_filters,
    4 * resnet_nature_comm.stem.num_filters,
    5 * resnet_nature_comm.stem.num_filters,
]
resnet_nature_comm.block.subsample_mode = "max"


# variant with attention of ResNet Nature Communications
resnet_nature_comm_se = deepcopy(resnet_nature_comm)
resnet_nature_comm_se.block = deepcopy(resnet_block_basic_se)

resnet_nature_comm_nl = deepcopy(resnet_nature_comm)
resnet_nature_comm_nl.block = deepcopy(resnet_block_basic_nl)

resnet_nature_comm_gc = deepcopy(resnet_nature_comm)
resnet_nature_comm_gc.block = deepcopy(resnet_block_basic_gc)

resnet_nature_comm_bottle_neck = deepcopy(resnet_nature_comm)
resnet_nature_comm_bottle_neck.building_block = "bottleneck"
resnet_nature_comm_bottle_neck.block = deepcopy(resnet_bottle_neck_B)

resnet_nature_comm_bottle_neck_se = deepcopy(resnet_nature_comm_bottle_neck)
resnet_nature_comm_bottle_neck_se.block = deepcopy(resnet_bottle_neck_se)

resnet_nature_comm_bottle_neck_gc = deepcopy(resnet_nature_comm_bottle_neck)
resnet_nature_comm_bottle_neck_gc.block = deepcopy(resnet_bottle_neck_gc)

resnet_nature_comm_bottle_neck_nl = deepcopy(resnet_nature_comm_bottle_neck)
resnet_nature_comm_bottle_neck_nl.block = deepcopy(resnet_bottle_neck_nl)


# TResNet
tresnet_common = CFG()
tresnet_common.fs = 500
tresnet_common.subsample_lengths = [
    1,
    2,
    2,
    2,
]
tresnet_common.filter_lengths = 11
tresnet_common.groups = 1
tresnet_common.increase_channels_method = "conv"
tresnet_common.stem = CFG()
tresnet_common.stem.subsample_mode = "s2d"
tresnet_common.stem.num_filters = 64
tresnet_common.stem.filter_lengths = 1
tresnet_common.stem.conv_stride = 1
tresnet_common.stem.pool_size = 1
tresnet_common.stem.pool_stride = 1
tresnet_common.kernel_initializer = "he_normal"
tresnet_common.kw_initializer = {}
tresnet_common.activation = "leaky"  # "mish", "swish"
tresnet_common.kw_activation = {"inplace": True}
tresnet_common.bias = False

tresnet_common.building_block = ["basic", "basic", "bottleneck", "bottleneck"]
tresnet_common.block = [
    deepcopy(resnet_block_basic_se),
    deepcopy(resnet_block_basic_se),
    deepcopy(resnet_bottle_neck_se),
    deepcopy(resnet_bottle_neck_B),
]
tresnet_common.block[0].reduction = 4
tresnet_common.block[1].reduction = 4
tresnet_common.block[2].pos = 2
for b in tresnet_common.block:
    b.kernel_initializer = tresnet_common.kernel_initializer
    b.kw_initializer = tresnet_common.kw_initializer
    b.activation = tresnet_common.activation
    b.kw_activation = tresnet_common.kw_activation
    b.bias = tresnet_common.bias
    b.subsample_mode = "blur"
    b.filt_size = 7  # for blur subsampling
    b.conv_type = "aa"

# TResNet-F, femto
tresnetF = deepcopy(tresnet_common)
tresnetF.filter_lengths = 17
tresnetF.num_blocks = [
    1,
    1,
    1,
    1,
]
tresnetF.stem.num_filters = 32
for b in tresnetF.block:
    b.conv_type = "separable"

# TResNet-P, pico
tresnetP = deepcopy(tresnet_common)
tresnetP.filter_lengths = 17
tresnetP.num_blocks = [
    1,
    1,
    1,
    1,
]
tresnetP.stem.num_filters = 56

# TResNet-N
tresnetN = deepcopy(tresnet_common)
tresnetN.filter_lengths = 15
tresnetN.num_blocks = [
    2,
    2,
    2,
    2,
]
tresnetN.stem.num_filters = 56

# TResNet-S
tresnetS = deepcopy(tresnet_common)
tresnetS.filter_lengths = 13
tresnetS.num_blocks = [
    3,
    4,
    6,
    3,
]
tresnetS.stem.num_filters = 56

# TResNet-M
tresnetM = deepcopy(tresnet_common)
tresnetM.num_blocks = [
    3,
    4,
    11,
    3,
]
tresnetM.stem.num_filters = 64

# TResNet-L
tresnetL = deepcopy(tresnet_common)
tresnetL.filter_lengths = [
    11,
    11,
    9,
    7,
]
tresnetL.num_blocks = [
    4,
    5,
    18,
    3,
]
tresnetL.stem.num_filters = 76

# TResNet-XL
tresnetXL = deepcopy(tresnet_common)
tresnetXL.filter_lengths = [
    11,
    9,
    9,
    7,
]
tresnetXL.num_blocks = [
    4,
    5,
    24,
    3,
]
tresnetXL.stem.num_filters = 84

# TResNet V2
tresnetM_V2 = deepcopy(tresnetM)
tresnetM_V2.building_block = ["bottleneck", "bottleneck", "bottleneck", "bottleneck"]
tresnetM_V2.block = [
    deepcopy(resnet_bottle_neck_se),
    deepcopy(resnet_bottle_neck_se),
    deepcopy(resnet_bottle_neck_se),
    deepcopy(resnet_bottle_neck_B),
]
tresnetM_V2.block[0].pos = 2
tresnetM_V2.block[1].pos = 2
tresnetM_V2.block[2].pos = 2



"""
"""

from typing import Any, Optional

import torch

from ..utils.utils_signal_t import resample as resample_t

__all__ = [
    "Resample",
]


class Resample(torch.nn.Module):
    """Resample the signal into fixed sampling frequency or length.

    Parameters
    ----------
    fs : int, optional
        Sampling frequency of the source signal to be resampled.
    dst_fs : int, optional
        Sampling frequency of the resampled ECG.
    siglen : int, optional
        Number of samples in the resampled ECG.
    inplace : bool, default False
        Whether to perform the resampling in-place.

    NOTE
    ----
    One and only one of `fs` and `siglen` should be set.
    If `fs` is set, `src_fs` should also be set.


    TODO
    ----
    Consider vectorized :func:`scipy.signal.resample`?

    """

    __name__ = "Resample"

    def __init__(
        self,
        fs: Optional[int] = None,
        dst_fs: Optional[int] = None,
        siglen: Optional[int] = None,
        inplace: bool = False,
        **kwargs: Any
    ) -> None:
        super().__init__()
        self.dst_fs = dst_fs
        self.fs = fs
        self.siglen = siglen
        self.inplace = inplace
        assert (
            sum([bool(self.fs), bool(self.siglen)]) == 1
        ), "one and only one of `fs` and `siglen` should be set"
        if self.dst_fs is not None:
            assert self.fs is not None, "if `dst_fs` is set, `fs` should also be set"
            self.scale_factor = self.dst_fs / self.fs

    def forward(self, sig: torch.Tensor) -> torch.Tensor:
        """Apply the resampling to the signal tensor.

        Parameters
        ----------
        sig : torch.Tensor
            The signal tensor to be resampled,
            of shape ``(..., n_leads, siglen)``.

        Returns
        -------
        torch.Tensor
            The resampled signal tensor,
            of shape ``(..., n_leads, siglen)``.

        """
        sig = resample_t(
            sig=sig,
            fs=self.fs,
            dst_fs=self.dst_fs,
            siglen=self.siglen,
            inplace=self.inplace,
        )
        return sig



"""
"""

from numbers import Real
from typing import Any, Optional

import torch

from .._preprocessors.base import preprocess_multi_lead_signal


__all__ = [
    "BandPass",
]


class BandPass(torch.nn.Module):
    """Bandpass filtering preprocessor.

    Parameters
    ----------
    fs : numbers.Real
        Sampling frequency of the ECG signal to be filtered.
    lowcut : numbers.Real, optional
        Low cutoff frequency.
    highcut : numbers.Real, optional
        High cutoff frequency.
    inplace : bool, default True
        Whether to perform the filtering in-place.
    kwargs : dict, optional
        Other keyword arguments for :class:`torch.nn.Module`.

    """

    __name__ = "BandPass"

    def __init__(
        self,
        fs: Real,
        lowcut: Optional[Real] = 0.5,
        highcut: Optional[Real] = 45,
        inplace: bool = True,
        **kwargs: Any
    ) -> None:
        super().__init__()
        self.fs = fs
        self.lowcut = lowcut
        self.highcut = highcut
        assert any(
            [self.lowcut is not None, self.highcut is not None]
        ), "At least one of lowcut and highcut should be set"
        if not self.lowcut:
            self.lowcut = 0
        if not self.highcut:
            self.highcut = float("inf")
        self.inplace = inplace

    def forward(self, sig: torch.Tensor) -> torch.Tensor:
        """Apply the preprocessor to the signal tensor.

        Parameters
        ----------
        sig : torch.Tensor
            The ECG signal tensor,
            of shape ``(batch, lead, siglen)``.

        Returns
        -------
        torch.Tensor
            The bandpass filtered ECG signal tensor,
            of shape ``(batch, lead, siglen)``.

        """
        if not self.inplace:
            sig = sig.clone()
        sig = torch.as_tensor(
            preprocess_multi_lead_signal(
                raw_sig=sig.cpu().numpy(),
                fs=self.fs,
                band_fs=[self.lowcut, self.highcut],
            ).copy(),
            dtype=sig.dtype,
            device=sig.device,
        )
        return sig



# -*- coding: utf-8 -*-

import time
from copy import deepcopy
from pathlib import Path
from typing import Any, Optional, Sequence, Dict, List, Union

import numpy as np
import pandas as pd
import h5py

from ...cfg import DEFAULTS
from ...utils.download import http_get
from ...utils.misc import get_record_list_recursive3, ms2samples, add_docstring
from ...utils import EAK
from ..base import DEFAULT_FIG_SIZE_PER_SEC, _DataBase, DataBaseInfo, _PlotCfg


__all__ = [
    "SPH",
]


_SPH_INFO = DataBaseInfo(
    title="""
    Shandong Provincial Hospital Database
    """,
    about=r"""
    1. contains 25770 ECG records from 24666 patients (55.36% male and 44.64% female), with between 10 and 60 seconds
    2. sampling frequency is 500 Hz
    3. records were acquired from Shandong Provincial Hospital (SPH) between 2019/08 and 2020/08
    4. diagnostic statements of all ECG records are in full compliance with the AHA/ACC/HRS recommendations, consisting of 44 primary statements and 15 modifiers
    5. 46.04% records in the dataset contain ECG abnormalities, and 14.45% records have multiple diagnostic statements
    6. (IMPORTANT) noises caused by the power line interference, baseline wander, and muscle contraction have been removed by the machine
    7. (Label production) The ECG analysis system automatically calculate nine ECG features for reference, which include heart rate, P wave duration, P-R interval, QRS duration, QT interval, corrected QT (QTc) interval, QRS axis, the amplitude of the R wave in lead V5 (RV5), and the amplitude of the S wave in lead V1 (SV1). A cardiologist made the final diagnosis in consideration of the patient health record.
    8. The paper [1]_, [2]_. Data can be downloaded from [3]_. The annotation system is described in [4]_.
    """,
    usage=[
        "ECG arrhythmia detection",
    ],
    references=[
        "https://www.nature.com/articles/s41597-022-01403-5",
        "Liu, H., Chen, D., Chen, D. et al. A large-scale multi-label 12-lead electrocardiogram database with standardized diagnostic statements. Sci Data 9, 272 (2022). https://doi.org/10.1038/s41597-022-01403-5",
        "https://springernature.figshare.com/collections/A_large-scale_multi-label_12-lead_electrocardiogram_database_with_standardized_diagnostic_statements/5779802/1",
        "Mason, J. W., Hancock, E. W. & Gettes, L. S. Recommendations for the standardization and interpretation of the electrocardiogram. Circulation 115, 1325–1332 (2007).",
    ],
    doi=[
        "10.1038/s41597-022-01403-5",
        "10.6084/m9.figshare.c.5779802.v1",
    ],
)


@add_docstring(_SPH_INFO.format_database_docstring(), mode="prepend")
class SPH(_DataBase):
    """
    Parameters
    ----------
    db_dir : str or pathlib.Path, optional
        Storage path of the database.
        If not specified, data will be fetched from Physionet.
    working_dir : str, optional
        Working directory, to store intermediate files and log files.
    verbose : int, default 1
        Level of logging verbosity.
    kwargs : dict, optional
        Auxilliary key word arguments

    """

    __name__ = "SPH"

    def __init__(
        self,
        db_dir: Optional[Union[str, Path]] = None,
        working_dir: Optional[Union[str, Path]] = None,
        verbose: int = 1,
        **kwargs: Any,
    ) -> None:
        super().__init__(
            db_name="SPH",
            db_dir=db_dir,
            working_dir=working_dir,
            verbose=verbose,
            **kwargs,
        )
        self.data_ext = "h5"
        self.ann_ext = None
        self.header_ext = None

        self.fs = 500
        self.all_leads = deepcopy(EAK.Standard12Leads)

        self._version = "v1"

        self._df_code = None
        self._df_metadata = None
        self._all_records = None
        self._ls_rec()

    def _ls_rec(self) -> None:
        """Find all records in the database directory
        and store them (path, metadata, etc.) in some private attributes.
        """
        record_list_fp = self.db_dir / "RECORDS"
        self._df_records = pd.DataFrame()
        write_file = False
        if record_list_fp.is_file():
            self._df_records["record"] = [
                item
                for item in record_list_fp.read_text().splitlines()
                if len(item) > 0
            ]
            if self._subsample is not None:
                size = min(
                    len(self._df_records),
                    max(1, int(round(self._subsample * len(self._df_records)))),
                )
                self._df_records = self._df_records.sample(
                    n=size, random_state=DEFAULTS.SEED, replace=False
                )
            self._df_records["path"] = self._df_records["record"].apply(
                lambda x: (self.db_dir / x).resolve()
            )
            self._df_records["record"] = self._df_records["path"].apply(
                lambda x: x.name
            )
            indices = self._df_records["path"].apply(lambda x: x.is_file())
            self._df_records = self._df_records.loc[indices]
        if len(self._df_records) == 0:
            write_file = True
            self.logger.info(
                "Please wait patiently to let the reader find "
                "all records of the database from local storage..."
            )
            start = time.time()
            record_pattern = "A[\\d]{5}\\.h5"
            self._df_records["path"] = get_record_list_recursive3(
                self.db_dir, record_pattern, relative=False
            )
            if self._subsample is not None:
                size = min(
                    len(self._df_records),
                    max(1, int(round(self._subsample * len(self._df_records)))),
                )
                self._df_records = self._df_records.sample(
                    n=size, random_state=DEFAULTS.SEED, replace=False
                )
            self._df_records["path"] = self._df_records["path"].apply(lambda x: Path(x))
            self.logger.info(f"Done in {time.time() - start:.3f} seconds!")
            self._df_records["record"] = self._df_records["path"].apply(
                lambda x: x.name
            )
        self._df_records.set_index("record", inplace=True)
        self._all_records = self._df_records.index.values.tolist()
        if write_file and self._subsample is None:
            record_list_fp.write_text(
                "\n".join(
                    self._df_records["path"]
                    .apply(lambda x: x.relative_to(self.db_dir).as_posix())
                    .tolist()
                )
            )

        if (self.db_dir / "code.csv").is_file():
            self._df_code = pd.read_csv(self.db_dir / "code.csv").astype(str)
        if (self.db_dir / "metadata.csv").is_file():
            self._df_metadata = pd.read_csv(self.db_dir / "metadata.csv")

    def get_subject_id(self, rec: Union[str, int]) -> str:
        """Attach a unique subject ID for the record.

        Parameters
        ----------
        rec : str or int
            Record name or index of the record in :attr:`all_records`

        Returns
        -------
        sid : str
            Subject ID associated with the record.

        """
        if isinstance(rec, int):
            rec = self[rec]
        sid = self._df_metadata.loc[self._df_metadata["ECG_ID"] == rec][
            "Patient_ID"
        ].iloc[0]
        return sid

    def load_data(
        self,
        rec: Union[str, int],
        leads: Optional[Union[str, int, List[Union[str, int]]]] = None,
        data_format: str = "channel_first",
        units: str = "mV",
    ) -> np.ndarray:
        """Load ECG data from h5 file of the record.

        Parameters
        ----------
        rec : str or int
            Record name or index of the record in :attr:`all_records`.
        leads : str or int or List[str] or List[int], optional
            The leads of the ECG data to be loaded.
        data_format : str, default "channel_first"
            Format of the ECG data,
            "channel_last" (alias "lead_last"), or
            "channel_first" (alias "lead_first").
        units : str, default "mV"
            Units of the output signal,
            can also be "μV" (alias "uV", "muV").

        Returns
        -------
        data : numpy.ndarray
            The loaded ECG data.

        """
        assert data_format.lower() in [
            "channel_first",
            "lead_first",
            "channel_last",
            "lead_last",
        ], f"Invalid data_format: `{data_format}`"

        _leads = self._normalize_leads(leads, numeric=True)

        with h5py.File(self.get_absolute_path(rec, extension=self.data_ext), "r") as f:
            data = f["ecg"][_leads].astype(DEFAULTS.DTYPE.NP)

        if units.lower() in ["uv", "μv", "muv"]:
            data = data * 1000
        elif units.lower() != "mv":
            raise AssertionError(f"Invalid units: `{units}`")

        if data_format.lower() in ["channel_last", "lead_last"]:
            data = data.T

        return data

    def load_ann(
        self, rec: Union[str, int], ann_format: str = "c", ignore_modifier: bool = True
    ) -> List[str]:
        """Load annotation from the metadata file.

        Parameters
        ----------
        rec : int or str
            Record name or index of the record in :attr:`all_records`.
        ann_format : str, default "a"
            Format of labels, one of the following (case insensitive):

                - "a": abbreviations
                - "f": full names
                - "c": AHACode

        ignore_modifier : bool, default True
            Whether to ignore the modifiers of the annotations or not.
            For example, "60+310" will be converted to "60"

        Returns
        -------
        labels : List[str]
            The list of labels.

        """
        if isinstance(rec, int):
            rec = self[rec]
        labels = [
            lb.strip()
            for lb in self._df_metadata[self._df_metadata["ECG_ID"] == rec]["AHA_Code"]
            .iloc[0]
            .split(";")
        ]
        modifiers = [lb.split("+")[1] if "+" in lb else "" for lb in labels]
        if ignore_modifier:
            labels = [lb.split("+")[0] for lb in labels]

        if ann_format.lower() == "c":
            pass  # default format
        elif ann_format.lower() == "f":
            labels = [
                self._df_code[self._df_code["Code"] == lb.split("+")[0]][
                    "Description"
                ].iloc[0]
                for lb in labels
            ]
            if not ignore_modifier:
                labels = [
                    f"""{self._df_code[self._df_code["Code"] == m]["Description"].iloc[0]} {lb}"""
                    if len(m) > 0
                    else lb
                    for lb, m in zip(labels, modifiers)
                ]
        elif ann_format.lower() == "a":
            raise NotImplementedError("Abbreviations are not supported yet")
        else:
            raise ValueError(f"Unknown annotation format: `{ann_format}`")
        return labels

    def get_subject_info(
        self, rec_or_sid: Union[str, int], items: Optional[List[str]] = None
    ) -> dict:
        """Read auxiliary information of a subject (a record)
        from the header files.

        Parameters
        ----------
        rec : int or str
            Record name, or index of the record in :attr:`all_records`,
            or the subject ID.
        items : List[str], optional
            Items of information to be returned (e.g. age, sex, etc.).

        Returns
        -------
        subject_info : dict
            Information about the subject, including
            "age", "sex".

        """
        if isinstance(rec_or_sid, int):
            rec_or_sid = self[rec_or_sid]
            row = self._df_metadata[self._df_metadata["ECG_ID"] == rec_or_sid].iloc[0]
        else:
            if rec_or_sid.startswith("A"):
                row = self._df_metadata[self._df_metadata["ECG_ID"] == rec_or_sid].iloc[
                    0
                ]
            else:
                row = self._df_metadata[
                    self._df_metadata["Patient_ID"] == rec_or_sid
                ].iloc[0]
        if items is None or len(items) == 0:
            info_items = [
                "age",
                "sex",
            ]
        else:
            info_items = items
        subject_info = {item: row[item.capitalize()] for item in info_items}

        return subject_info

    def get_age(self, rec: Union[str, int]) -> int:
        """Get the age of the subject that the record belongs to.

        Parameters
        ----------
        rec : int or str
            Record name or index of the record in :attr:`all_records`.

        Returns
        -------
        age : int
            Age of the subject.

        """
        if isinstance(rec, int):
            rec = self[rec]
        age = (
            self._df_metadata[self._df_metadata["ECG_ID"] == rec]["Age"].iloc[0].item()
        )
        return age

    def get_sex(self, rec: Union[str, int]) -> str:
        """Get the sex of the subject that the record belongs to.

        Parameters
        ----------
        rec : int or str
            Record name or index of the record in :attr:`all_records`.

        Returns
        -------
        sex : str
            Sex of the subject.

        """
        if isinstance(rec, int):
            rec = self[rec]
        sex = self._df_metadata[self._df_metadata["ECG_ID"] == rec]["Sex"].iloc[0]
        return sex

    def get_siglen(self, rec: Union[str, int]) -> int:
        """Get the length of the ECG signal of the record.

        Parameters
        ----------
        rec : int or str
            Record name or index of the record in :attr:`all_records`.

        Returns
        -------
        siglen : int
            Length of the ECG signal of the record.

        """
        if isinstance(rec, int):
            rec = self[rec]
        siglen = (
            self._df_metadata[self._df_metadata["ECG_ID"] == rec]["N"].iloc[0].item()
        )
        return siglen

    @property
    def url(self) -> Dict[str, str]:
        return {
            "metadata.csv": "https://springernature.figshare.com/ndownloader/files/34793152",
            "code.csv": "https://springernature.figshare.com/ndownloader/files/32630954",
            "records.tar": "https://springernature.figshare.com/ndownloader/files/32630684",
        }

    def download(self, files: Optional[Union[str, Sequence[str]]]) -> None:
        """Download the database from the figshare website."""
        if files is None:
            files = self.url.keys()
        if isinstance(files, str):
            files = [files]
        assert set(files).issubset(
            self.url
        ), f"`files` should be a subset of {list(self.url)}"
        for filename in files:
            url = self.url[filename]
            if not (self.db_dir / filename).is_file():
                http_get(url, self.db_dir, filename=filename)
        self._ls_rec()

    def plot(
        self,
        rec: Union[str, int],
        data: Optional[np.ndarray] = None,
        ann: Optional[Sequence[str]] = None,
        ticks_granularity: int = 0,
        leads: Optional[Union[str, int, List[Union[str, int]]]] = None,
        same_range: bool = False,
        waves: Optional[Dict[str, Sequence[int]]] = None,
        **kwargs: Any,
    ) -> None:
        """
        Plot the signals of a record or external signals (units in μV),
        with metadata (fs, labels, tranche, etc.),
        possibly also along with wave delineations.

        Parameters
        ----------
        rec : int or str
            Record name or index of the record in :attr:`all_records`.
        data : numpy.ndarray, optional
            (12-lead) ECG signal to plot.
            Should be of the format "channel_first",
            and compatible with `leads`.
            If not None, data of `rec` will not be used.
            Tthis is useful when plotting filtered data.
        ann : Sequence[str], optional
            Annotations for `data`.
            Ignored if `data` is None.
        ticks_granularity : int, default 0
            Granularity to plot axis ticks, the higher the more ticks.
            0 (no ticks) --> 1 (major ticks) --> 2 (major + minor ticks)
        leads : str or int or List[str] or List[int], optional
            The leads of the ECG signal to plot.
        same_range : bool, default False
            If True, forces all leads to have the same y range.
        waves : dict, optional
            A dictionary containing the
            indices of the wave critical points, including
            "p_onsets", "p_peaks", "p_offsets",
            "q_onsets", "q_peaks", "r_peaks", "s_peaks", "s_offsets",
            "t_onsets", "t_peaks", "t_offsets".
        kwargs : dict, optional
            Additional keyword arguments to pass to :func:`matplotlib.pyplot.plot`.

        TODO
        ----
        1. Slice too long records, and plot separately for each segment.
        2. Plot waves using :func:`matplotlib.pyplot.axvspan`.

        NOTE
        ----
        `Locator` of ``plt`` has default `MAXTICKS` of 1000.
        If not modifying this number,
        at most 40 seconds of signal could be plotted once.

        Contributors: Jeethan, and WEN Hao

        """
        if isinstance(rec, int):
            rec = self[rec]

        if "plt" not in dir():
            import matplotlib.pyplot as plt

            plt.MultipleLocator.MAXTICKS = 3000

        _leads = self._normalize_leads(leads, numeric=False)
        lead_indices = [self.all_leads.index(ld) for ld in _leads]

        if data is None:
            _data = self.load_data(rec, data_format="channel_first", units="μV")[
                lead_indices
            ]
        else:
            units = self._auto_infer_units(data)
            self.logger.info(f"input data is auto detected to have units in `{units}`")
            if units.lower() == "mv":
                _data = 1000 * data
            else:
                _data = data
            assert _data.shape[0] == len(_leads), (
                f"number of leads from data of shape ({_data.shape[0]}) does not "
                f"match the length ({len(_leads)}) of `leads`"
            )

        if same_range:
            y_ranges = np.ones((_data.shape[0],)) * np.max(np.abs(_data)) + 100
        else:
            y_ranges = np.max(np.abs(_data), axis=1) + 100

        if waves:
            if waves.get("p_onsets", None) and waves.get("p_offsets", None):
                p_waves = [
                    [onset, offset]
                    for onset, offset in zip(waves["p_onsets"], waves["p_offsets"])
                ]
            elif waves.get("p_peaks", None):
                p_waves = [
                    [
                        max(0, p + ms2samples(_PlotCfg.p_onset, fs=self.fs)),
                        min(
                            _data.shape[1],
                            p + ms2samples(_PlotCfg.p_offset, fs=self.fs),
                        ),
                    ]
                    for p in waves["p_peaks"]
                ]
            else:
                p_waves = []
            if waves.get("q_onsets", None) and waves.get("s_offsets", None):
                qrs = [
                    [onset, offset]
                    for onset, offset in zip(waves["q_onsets"], waves["s_offsets"])
                ]
            elif waves.get("q_peaks", None) and waves.get("s_peaks", None):
                qrs = [
                    [
                        max(0, q + ms2samples(_PlotCfg.q_onset, fs=self.fs)),
                        min(
                            _data.shape[1],
                            s + ms2samples(_PlotCfg.s_offset, fs=self.fs),
                        ),
                    ]
                    for q, s in zip(waves["q_peaks"], waves["s_peaks"])
                ]
            elif waves.get("r_peaks", None):
                qrs = [
                    [
                        max(0, r + ms2samples(_PlotCfg.qrs_radius, fs=self.fs)),
                        min(
                            _data.shape[1],
                            r + ms2samples(_PlotCfg.qrs_radius, fs=self.fs),
                        ),
                    ]
                    for r in waves["r_peaks"]
                ]
            else:
                qrs = []
            if waves.get("t_onsets", None) and waves.get("t_offsets", None):
                t_waves = [
                    [onset, offset]
                    for onset, offset in zip(waves["t_onsets"], waves["t_offsets"])
                ]
            elif waves.get("t_peaks", None):
                t_waves = [
                    [
                        max(0, t + ms2samples(_PlotCfg.t_onset, fs=self.fs)),
                        min(
                            _data.shape[1],
                            t + ms2samples(_PlotCfg.t_offset, fs=self.fs),
                        ),
                    ]
                    for t in waves["t_peaks"]
                ]
            else:
                t_waves = []
        else:
            p_waves, qrs, t_waves = [], [], []
        palette = {
            "p_waves": "green",
            "qrs": "red",
            "t_waves": "yellow",
        }
        plot_alpha = 0.4

        if ann is None or data is None:
            ann = self.load_ann(rec, ann_format="f")

        nb_leads = len(_leads)

        seg_len = self.fs * 25  # 25 seconds
        nb_segs = _data.shape[1] // seg_len

        t = np.arange(_data.shape[1]) / self.fs
        duration = len(t) / self.fs
        fig_sz_w = int(round(DEFAULT_FIG_SIZE_PER_SEC * duration))
        fig_sz_h = 6 * np.maximum(y_ranges, 750) / 1500
        fig, axes = plt.subplots(
            nb_leads, 1, sharex=False, figsize=(fig_sz_w, np.sum(fig_sz_h))
        )
        if nb_leads == 1:
            axes = [axes]
        for idx in range(nb_leads):
            axes[idx].plot(
                t,
                _data[idx],
                color="black",
                linewidth="2.0",
                label=f"lead - {_leads[idx]}",
            )
            axes[idx].axhline(y=0, linestyle="-", linewidth="1.0", color="red")
            # NOTE that `Locator` has default `MAXTICKS` equal to 1000
            if ticks_granularity >= 1:
                axes[idx].xaxis.set_major_locator(plt.MultipleLocator(0.2))
                axes[idx].yaxis.set_major_locator(plt.MultipleLocator(500))
                axes[idx].grid(
                    which="major", linestyle="-", linewidth="0.4", color="red"
                )
            if ticks_granularity >= 2:
                axes[idx].xaxis.set_minor_locator(plt.MultipleLocator(0.04))
                axes[idx].yaxis.set_minor_locator(plt.MultipleLocator(100))
                axes[idx].grid(
                    which="minor", linestyle=":", linewidth="0.2", color="gray"
                )
            # add extra info. to legend
            # https://stackoverflow.com/questions/16826711/is-it-possible-to-add-a-string-as-a-legend-item-in-matplotlib
            axes[idx].plot([], [], " ", label=f"labels - {','.join(ann)}")
            for w in ["p_waves", "qrs", "t_waves"]:
                for itv in eval(w):
                    axes[idx].axvspan(
                        t[itv[0]], t[itv[1]], color=palette[w], alpha=plot_alpha
                    )
            axes[idx].legend(loc="upper left", fontsize=14)
            axes[idx].set_xlim(t[0], t[-1])
            axes[idx].set_ylim(min(-600, -y_ranges[idx]), max(600, y_ranges[idx]))
            axes[idx].set_xlabel("Time [s]", fontsize=16)
            axes[idx].set_ylabel("Voltage [μV]", fontsize=16)
        plt.subplots_adjust(hspace=0.05)
        fig.tight_layout()
        if kwargs.get("save_path", None):
            plt.savefig(kwargs["save_path"], dpi=200, bbox_inches="tight")
        else:
            plt.show()

    @property
    def database_info(self) -> DataBaseInfo:
        return _SPH_INFO



"""
loggers, including (planned) CSVLogger, TensorBoardXLogger, WandbLogger

with reference to `loggers` of `textattack` and `loggers` of `pytorch-lightning`

"""

import csv
import logging
from abc import ABC, abstractmethod
from datetime import datetime
from numbers import Real
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

import pandas as pd
import tensorboardX
import torch

from ..cfg import DEFAULTS
from ..utils.misc import ReprMixin, get_date_str, init_logger, add_docstring


__all__ = [
    "BaseLogger",
    "TxtLogger",
    "CSVLogger",
    "TensorBoardXLogger",
    # "WandbLogger",
    "LoggerManager",
]


_log_metrics_doc = """Log metrics.

        Parameters
        ----------
        metrics : dict
            The metrics to be logged.
        step : int, optional
            The number of (global) steps of training.
        epoch: int, optional,
            The epoch number of training.
        part : str, optional
            The part of the training data the metrics computed from,
            can be "train" or "val" or "test", etc.

        """

_log_message_doc = """Log a message.

        Parameters
        ----------
        msg : str
            The message to be logged.
        level : {logging.DEBUG, logging.INFO, logging.WARNING, logging.ERROR, logging.CRITICAL}, optional
            The level of the message.

        """

_epoch_start_doc = """Message logged at the start of each epoch.

        Parameters
        ----------
        epoch : int
            The number of the epoch.

        """

_epoch_end_doc = """Actions to be performed at the end of each epoch.

        Parameters
        ----------
        epoch : int
            The number of the epoch.

        """


class BaseLogger(ReprMixin, ABC):
    """The abstract base class of all loggers."""

    __name__ = "BaseLogger"

    @add_docstring(_log_metrics_doc)
    @abstractmethod
    def log_metrics(
        self,
        metrics: Dict[str, Union[Real, torch.Tensor]],
        step: Optional[int] = None,
        epoch: Optional[int] = None,
        part: str = "train",
    ) -> None:
        raise NotImplementedError

    @add_docstring(_log_message_doc)
    @abstractmethod
    def log_message(self, msg: str, level: int = logging.INFO) -> None:
        raise NotImplementedError

    @abstractmethod
    def flush(self) -> None:
        """Flush the logger."""
        raise NotImplementedError

    @abstractmethod
    def close(self) -> None:
        """Close the logger."""
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def from_config(cls, config: Dict[str, Any]) -> Any:
        """Initialize the logger from a config dict."""
        raise NotImplementedError

    @add_docstring(_epoch_start_doc)
    def epoch_start(self, epoch: int) -> None:
        pass

    @add_docstring(_epoch_end_doc)
    def epoch_end(self, epoch: int) -> None:
        pass

    @property
    def log_dir(self) -> str:
        """Directory to save the log file."""
        return self._log_dir

    @property
    @abstractmethod
    def filename(self) -> str:
        """Filename of the log file."""
        raise NotImplementedError

    def extra_repr_keys(self) -> List[str]:
        return super().extra_repr_keys() + [
            "filename",
        ]


class TxtLogger(BaseLogger):
    """Logger that logs to a text file.

    Parameters
    ----------
    log_dir : str or path.Path, optional
        The directory to save the log file.
    log_suffix : str, optional
        The suffix of the log file.

    """

    __name__ = "TxtLogger"

    def __init__(
        self,
        log_dir: Optional[Union[str, Path]] = None,
        log_suffix: Optional[str] = None,
    ) -> None:
        self._log_dir = Path(log_dir or DEFAULTS.log_dir)
        self._log_dir.mkdir(parents=True, exist_ok=True)
        if log_suffix is None:
            log_suffix = ""
        else:
            log_suffix = f"_{log_suffix}"
        self.log_file = f"{DEFAULTS.prefix}_{get_date_str()}{log_suffix}.txt"
        self.logger = init_logger(self.log_dir, self.log_file, verbose=1)
        self.step = -1

    @add_docstring(_log_metrics_doc)
    def log_metrics(
        self,
        metrics: Dict[str, Union[Real, torch.Tensor]],
        step: Optional[int] = None,
        epoch: Optional[int] = None,
        part: str = "train",
    ) -> None:
        if step is not None:
            self.step = step
        else:
            self.step += 1
        prefix = f"Step {step}: "
        if epoch is not None:
            prefix = f"Epoch {epoch} / {prefix}"
        _metrics = {
            k: v.item() if isinstance(v, torch.Tensor) else v
            for k, v in metrics.items()
        }
        spaces = len(max(_metrics.keys(), key=len))
        msg = (
            f"{part.capitalize()} Metrics:\n{self.short_sep}\n"
            + "\n".join(
                [
                    f"{prefix}{part}/{k} : {' '*(spaces-len(k))}{v:.4f}"
                    for k, v in _metrics.items()
                ]
            )
            + f"\n{self.short_sep}"
        )
        self.log_message(msg)

    @add_docstring(_log_message_doc)
    def log_message(self, msg: str, level: int = logging.INFO) -> None:
        self.logger.log(level, msg)

    @property
    def long_sep(self) -> str:
        return "-" * 110

    @property
    def short_sep(self) -> str:
        return "-" * 50

    @add_docstring(_epoch_start_doc)
    def epoch_start(self, epoch: int) -> None:
        self.logger.info(f"Train epoch_{epoch}:\n{self.long_sep}")

    @add_docstring(_epoch_end_doc)
    def epoch_end(self, epoch: int) -> None:
        self.logger.info(f"{self.long_sep}\n")

    def flush(self) -> None:
        """Flush the log file."""
        for h in self.logger.handlers:
            if hasattr(h, "flush"):
                h.flush()

    def close(self) -> None:
        """Close the log file."""
        for h in self.logger.handlers:
            h.close()
            self.logger.removeHandler(h)
        logging.shutdown()

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> "TxtLogger":
        """Create a logger from a config.

        Parameters
        ----------
        config : dict
            The config to create the logger.

        Returns
        -------
        TxtLogger
            The instance of the created logger.

        """
        return cls(config.get("log_dir", None), config.get("log_suffix", None))

    @property
    def filename(self) -> str:
        """Filename of the log file."""
        return str(self.log_dir / self.log_file)


class CSVLogger(BaseLogger):
    """Logger that logs to a CSV file.

    Parameters
    ----------
    log_dir : str or path.Path, optional
        The directory to save the log file.
    log_suffix : str, optional
        The suffix of the log file.

    """

    __name__ = "CSVLogger"

    def __init__(
        self,
        log_dir: Optional[Union[str, Path]] = None,
        log_suffix: Optional[str] = None,
    ) -> None:
        self._log_dir = Path(log_dir or DEFAULTS.log_dir)
        self._log_dir.mkdir(parents=True, exist_ok=True)
        if log_suffix is None:
            log_suffix = ""
        else:
            log_suffix = f"_{log_suffix}"
        self.log_file = f"{DEFAULTS.prefix}_{get_date_str()}{log_suffix}.csv"
        self.logger = pd.DataFrame()
        self.step = -1
        self._flushed = True

    @add_docstring(_log_metrics_doc)
    def log_metrics(
        self,
        metrics: Dict[str, Union[Real, torch.Tensor]],
        step: Optional[int] = None,
        epoch: Optional[int] = None,
        part: str = "train",
    ) -> None:
        if step is not None:
            self.step = step
        else:
            self.step += 1
        row = {"step": self.step, "time": datetime.now(), "part": part}
        if epoch is not None:
            row.update({"epoch": epoch})
        row.update(
            {
                k: v.item() if isinstance(v, torch.Tensor) else v
                for k, v in metrics.items()
            }
        )
        # self.logger = self.logger.append(row, ignore_index=True)
        self.logger = pd.concat(
            [self.logger, pd.DataFrame(row, index=[0])], ignore_index=True
        )

        self._flushed = False

    def log_message(self, msg: str, level: int = logging.INFO) -> None:
        pass

    def flush(self) -> None:
        """Flush the log file."""
        if not self._flushed:
            self.logger.to_csv(self.filename, quoting=csv.QUOTE_NONNUMERIC, index=False)
            self._flushed = True

    def close(self) -> None:
        """Close the log file."""
        self.flush()

    def __del__(self):
        self.flush()
        del self

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> "CSVLogger":
        """Create a logger from a config.

        Parameters
        ----------
        config : dict
            The config to create the logger.

        Returns
        -------
        CSVLogger
            The instance of the created logger.

        """
        return cls(config.get("log_dir", None), config.get("log_suffix", None))

    @property
    def filename(self) -> str:
        """Filename of the log file."""
        return str(self.log_dir / self.log_file)


class TensorBoardXLogger(BaseLogger):
    """Logger that logs to a TensorBoardX file.

    Parameters
    ----------
    log_dir : str or path.Path, optional
        The directory to save the log file.
    log_suffix : str, optional
        The suffix of the log file.

    """

    __name__ = "TensorBoardXLogger"

    def __init__(
        self,
        log_dir: Optional[Union[str, Path]] = None,
        log_suffix: Optional[str] = None,
    ) -> None:
        self._log_dir = Path(log_dir or DEFAULTS.log_dir)
        self._log_dir.mkdir(parents=True, exist_ok=True)
        self.logger = tensorboardX.SummaryWriter(
            str(self._log_dir), filename_suffix=log_suffix or ""
        )
        self.log_file = self.logger.file_writer.event_writer._ev_writer._file_name
        self.step = -1

    @add_docstring(_log_metrics_doc)
    def log_metrics(
        self,
        metrics: Dict[str, Union[Real, torch.Tensor]],
        step: Optional[int] = None,
        epoch: Optional[int] = None,
        part: str = "train",
    ) -> None:
        if step is not None:
            self.step = step
        else:
            self.step += 1
        for k, v in metrics.items():
            if isinstance(v, torch.Tensor):
                v = v.item()
            self.logger.add_scalar(f"{part}/{k}", v, self.step)

    def log_message(self, msg: str, level: int = logging.INFO) -> None:
        pass

    def flush(self) -> None:
        """Flush the log file."""
        self.logger.flush()

    def close(self) -> None:
        """Close the log file."""
        self.logger.close()

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> "TensorBoardXLogger":
        """Create a logger from a config.

        Parameters
        ----------
        config : dict
            The config to create the logger.

        Returns
        -------
        TensorBoardXLogger
            The instance of the created logger.

        """
        return cls(config.get("log_dir", None), config.get("log_suffix", None))

    @property
    def filename(self) -> str:
        """Filename of the log file."""
        return str(self.log_dir / self.log_file)


# class WandbLogger(BaseLogger):
#     """ """

#     __name__ = "WandbLogger"

#     def __init__(
#         self,
#         log_dir: Optional[Union[str, Path]] = None,
#         log_suffix: Optional[str] = None,
#         project: Optional[str] = None,
#         entity: Optional[str] = None,
#         hyperparameters: Optional[dict] = None,
#     ) -> None:
#         """
#         to write docstring
#         """
#         self.__wandb = importlib.import_module("wandb")
#         self._log_dir = Path(log_dir or DEFAULTS.log_dir)
#         self._log_dir.mkdir(parents=True, exist_ok=True)
#         self._log_suffix = log_suffix
#         self._project = project
#         self._entity = entity
#         self._hyperparameters = hyperparameters
#         self.__wandb.init(
#             name=self._log_suffix,
#             dir=self._log_dir,
#             config=self._hyperparameters,
#             project=self._project,
#             entity=self._entity,
#         )

#     def log_metrics(
#         self, metrics: Dict[str, float], step: Optional[int] = None
#     ) -> None:
#         """ """
#         self.__wandb.log(metrics, step=step)

#     def log_message(self, msg: str, level: int = logging.INFO) -> None:
#         pass

#     def flush(self) -> None:
#         """ """
#         pass

#     def close(self) -> None:
#         """ """
#         self.__wandb.finish()

#     @classmethod
#     def from_config(cls, config: Dict[str, Any]) -> "WandbLogger":
#         """ """
#         return cls(
#             config.get("log_dir", None),
#             config.get("log_suffix", None),
#             config.get("project", None),
#             config.get("entity", None),
#             config.get("hyperparameters", None),
#         )

#     @property
#     def filename(self) -> str:
#         """ """
#         return self.__wandb.run.dir


class LoggerManager(ReprMixin):
    """Manager of loggers.

    This class manages multiple loggers and provides a unified interface to
    log metrics and messages.

    Parameters
    ----------
    log_dir : str or path.Path, optional
        The directory to save the log file.
    log_suffix : str, optional
        The suffix of the log file.

    """

    __name__ = "LoggerManager"

    def __init__(
        self,
        log_dir: Optional[Union[str, Path]] = None,
        log_suffix: Optional[str] = None,
    ) -> None:
        self._log_dir = Path(log_dir or DEFAULTS.log_dir)
        self._log_suffix = log_suffix
        self._loggers = []

    def _add_txt_logger(self) -> None:
        """Add a text logger to the manager."""
        self.loggers.append(TxtLogger(self._log_dir, self._log_suffix))

    def _add_csv_logger(self) -> None:
        """Add a csv logger to the manager."""
        self.loggers.append(CSVLogger(self._log_dir, self._log_suffix))

    def _add_tensorboardx_logger(self) -> None:
        """Add a tensorboardx logger to the manager."""
        self.loggers.append(TensorBoardXLogger(self._log_dir, self._log_suffix))

    # def _add_wandb_logger(self, **kwargs: dict) -> None:
    #     """ """
    #     raise NotImplementedError("NOT tested yet!")
    # self.loggers.append(WandbLogger(self._log_dir, self._log_suffix, **kwargs))

    @add_docstring(_log_metrics_doc)
    def log_metrics(
        self,
        metrics: Dict[str, Union[Real, torch.Tensor]],
        step: Optional[int] = None,
        epoch: Optional[int] = None,
        part: str = "train",
    ) -> None:
        for lgs in self.loggers:
            lgs.log_metrics(metrics, step, epoch, part)

    @add_docstring(_log_message_doc)
    def log_message(self, msg: str, level: int = logging.INFO) -> None:
        for lgs in self.loggers:
            lgs.log_message(msg, level)

    @add_docstring(_epoch_start_doc)
    def epoch_start(self, epoch: int) -> None:
        for lgs in self.loggers:
            lgs.epoch_start(epoch)

    @add_docstring(_epoch_end_doc)
    def epoch_end(self, epoch: int) -> None:
        for lgs in self.loggers:
            lgs.epoch_end(epoch)

    def flush(self) -> None:
        """Flush the loggers."""
        for lgs in self.loggers:
            lgs.flush()

    def close(self) -> None:
        """Close the loggers."""
        for lgs in self.loggers:
            lgs.close()

    @property
    def loggers(self) -> List[BaseLogger]:
        """List of loggers."""
        return self._loggers

    @property
    def log_dir(self) -> str:
        """Directory to save the log file."""
        return self._log_dir

    @property
    def log_suffix(self) -> str:
        """Suffix of the log file."""
        return self._log_suffix

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> "LoggerManager":
        """Create a logger manager from a configuration.

        Parameters
        ----------
        config : dict
            The configuration of the logger manager.

        Returns
        -------
        LoggerManager
            The instance of the created logger manager.

        """
        lm = cls(config.get("log_dir", None), config.get("log_suffix", None))
        if config.get("txt_logger", True):
            lm._add_txt_logger()
        if config.get("csv_logger", True):
            lm._add_csv_logger()
        if config.get("tensorboardx_logger", True):
            lm._add_tensorboardx_logger()
        # if config.get("wandb_logger", False):
        #     kwargs = config.get("wandb_logger", {})
        #     lm._add_wandb_logger(**kwargs)
        return lm

    def extra_repr_keys(self) -> List[str]:
        return super().extra_repr_keys() + [
            "loggers",
        ]



"""
Utilities for signal processing,
including spatial, temporal, spatio-temporal domains.
"""

import warnings
from copy import deepcopy
from numbers import Real
from typing import Iterable, Optional, Sequence, Tuple, Union

import numpy as np
from scipy import interpolate
from scipy.signal import butter, filtfilt, peak_prominences

from .utils_data import ensure_siglen


__all__ = [
    "smooth",
    "resample_irregular_timeseries",
    "detect_peaks",
    "remove_spikes_naive",
    "butter_bandpass_filter",
    "get_ampl",
    "normalize",
]


def smooth(
    x: np.ndarray,
    window_len: int = 11,
    window: str = "hanning",
    mode: str = "valid",
    keep_dtype: bool = True,
) -> np.ndarray:
    """Smooth the 1d data using a window with requested size.

    This method is originally from [#smooth]_,
    based on the convolution of a scaled window with the signal.
    The signal is prepared by introducing reflected copies of the signal
    (with the window size) in both ends so that transient parts are minimized
    in the begining and end part of the output signal.

    Parameters
    ----------
    x : numpy.ndarray
        The input signal.
    window_len : int, default 11
        Length of the smoothing window,
        (previously should be an odd integer,
        currently can be any (positive) integer).
    window : {"flat", "hanning", "hamming", "bartlett", "blackman"}, optional
        Type of window from, by default "hanning".
        See also :func:`numpy.hanning`, :func:`numpy.hamming`, etc.
        Flat type window will produce a moving average smoothing.
    mode : str, default "valid"
        Mode of convolution, see :func:`numpy.convolve` for details.
    keep_dtype : bool, default True
        Whether `dtype` of the returned value keeps
        the same with that of `x` or not.

    Returns
    -------
    y : numpy.ndarray
        The smoothed signal.

    Examples
    --------
    .. code-block:: python

        t = np.linspace(-2, 2, 50)
        x = np.sin(t) + np.random.randn(len(t)) * 0.1
        y = smooth(x)

    See also
    --------
    :func:`numpy.hanning`, :func:`numpy.hamming`,
    :func:`numpy.bartlett`, :func:`numpy.blackman`, :func:`numpy.convolve`,
    :func:`scipy.signal.lfilter`.

    TODO
    ----
    The window parameter could be the window itself
    if an array instead of a string.

    NOTE
    ----
    length(output) != length(input), to correct this, using

    .. code-block:: python

        return y[(window_len/2-1):-(window_len/2)]

    instead of just returning `y`.

    References
    ----------
    .. [#smooth] https://scipy-cookbook.readthedocs.io/items/SignalSmooth.html

    """
    radius = min(len(x), window_len)
    radius = radius if radius % 2 == 1 else radius - 1

    if x.ndim != 1:
        raise ValueError("function `smooth` only accepts 1 dimension arrays.")

    # if x.size < radius:
    #     raise ValueError("Input vector needs to be bigger than window size.")

    if radius < 3:
        return x

    if window not in ["flat", "hanning", "hamming", "bartlett", "blackman"]:
        raise ValueError(
            """ `window` should be of "flat", "hanning", "hamming", "bartlett", "blackman" """
        )

    s = np.r_[x[radius - 1 : 0 : -1], x, x[-2 : -radius - 1 : -1]]
    # print(len(s))
    if window == "flat":  # moving average
        w = np.ones(radius, "d")
    else:
        w = eval("np." + window + "(radius)")

    y = np.convolve(w / w.sum(), s, mode=mode)
    y = y[(radius // 2 - 1) : -(radius // 2) - 1]
    assert len(x) == len(y)

    if keep_dtype:
        y = y.astype(x.dtype)

    return y


def resample_irregular_timeseries(
    sig: np.ndarray,
    output_fs: Optional[Real] = None,
    method: str = "spline",
    return_with_time: bool = False,
    tnew: Optional[np.ndarray] = None,
    interp_kw: dict = {},
    verbose: int = 0,
) -> np.ndarray:
    """
    Resample the 2d irregular timeseries `sig` into a 1d or 2d
    regular time series with frequency `output_fs`,
    elements of `sig` are in the form ``[time, value]``,
    where the unit of `time` is ms.

    Parameters
    ----------
    sig : numpy.ndarray
        The 2d irregular timeseries.
        Each row is ``[time, value]``.
    output_fs : numbers.Real, optional
        the frequency of the output 1d regular timeseries,
        one and only one of `output_fs` and `tnew` should be specified
    method : str, default "spline"
        interpolation method, can be "spline" or "interp1d"
    return_with_time : bool, default False
        return a 2d array, with the 0-th coordinate being time
    tnew : array_like, optional
        the array of time of the output array,
        one and only one of `output_fs` and `tnew` should be specified
    interp_kw : dict, optional
        additional options for the corresponding methods in scipy.interpolate

    Returns
    -------
    numpy.ndarray
        A 1d or 2d regular time series with frequency `output_freq`.

    Examples
    --------
    .. code-block:: python

        fs = 100
        t_irr = np.sort(np.random.rand(fs)) * 1000
        vals = np.random.randn(fs)
        sig = np.stack([t_irr, vals], axis=1)
        sig_reg = resample_irregular_timeseries(sig, output_fs=fs * 2, return_with_time=True)
        sig_reg = resample_irregular_timeseries(sig, output_fs=fs, method="interp1d")
        t_irr_2 = np.sort(np.random.rand(2 * fs)) * 1000
        sig_reg = resample_irregular_timeseries(sig, tnew=t_irr_2, return_with_time=True)

    NOTE
    ----
    ``pandas`` also has the function to regularly resample irregular timeseries.

    """
    assert sig.ndim == 2, "`sig` should be a 2D array"
    assert method.lower() in [
        "spline",
        "interp1d",
    ], f"method `{method}` not supported"
    assert (
        sum([output_fs is None, tnew is None]) == 1
    ), "one and only one of `output_fs` and `tnew` should be specified"

    _interp_kw = deepcopy(interp_kw)

    if verbose >= 1:
        print(f"len(sig) = {len(sig)}")

    if len(sig) == 0:
        return np.array([])

    dtype = sig.dtype
    time_series = np.atleast_2d(sig).astype(dtype)
    if tnew is None:
        step_ts = 1000 / output_fs
        tot_len = int((time_series[-1][0] - time_series[0][0]) / step_ts) + 1
        xnew = time_series[0][0] + np.arange(0, tot_len * step_ts, step_ts)
    else:
        assert tnew.ndim == 1, "`tnew` should be a 1D array"
        xnew = np.array(tnew)
        tot_len = len(xnew)

    if verbose >= 1:
        print(
            f"time_series start ts = {time_series[0][0]}, end ts = {time_series[-1][0]}"
        )
        print(f"tot_len = {tot_len}")
        print(f"xnew start = {xnew[0]}, end = {xnew[-1]}")

    if method.lower() == "spline":
        m = len(time_series)
        w = interp_kw.get("w", np.ones(shape=(m,)))
        # s = interp_kw.get("s", np.random.uniform(m-np.sqrt(2*m),m+np.sqrt(2*m)))
        s = interp_kw.get("s", m - np.sqrt(2 * m))
        _interp_kw.update(w=w, s=s)

        tck = interpolate.splrep(time_series[:, 0], time_series[:, 1], **_interp_kw)

        regular_timeseries = interpolate.splev(xnew, tck)
    elif method.lower() == "interp1d":
        f = interpolate.interp1d(time_series[:, 0], time_series[:, 1], **_interp_kw)

        regular_timeseries = f(xnew)

    if return_with_time:
        return np.column_stack((xnew, regular_timeseries)).astype(dtype)
    else:
        return regular_timeseries.astype(dtype)


def detect_peaks(
    x: Sequence,
    mph: Optional[Real] = None,
    mpd: int = 1,
    threshold: Real = 0,
    left_threshold: Real = 0,
    right_threshold: Real = 0,
    prominence: Optional[Real] = None,
    prominence_wlen: Optional[int] = None,
    edge: Union[str, None] = "rising",
    kpsh: bool = False,
    valley: bool = False,
    show: bool = False,
    ax=None,
    verbose: int = 0,
) -> np.ndarray:
    """Detect peaks in data based on their amplitude and other features.

    Parameters
    ----------
    x : array_like
        1D array of data.
    mph : positive number, optional
        abbr. for maximum (minimum) peak height,
        detect peaks that are greater than minimum peak height (if parameter `valley` is False),
        or peaks that are smaller than maximum peak height (if parameter `valley` is True)
    mpd : positive integer, default 1
        abbr. for minimum peak distance,
        detect peaks that are at least separated by minimum peak distance (in number of samples)
    threshold : positive number, default 0
        detect peaks (valleys) that are greater (smaller) than `threshold`,
        in relation to their neighbors within the range of `mpd`
    left_threshold : positive number, default 0
        `threshold` that is restricted to the left
    right_threshold : positive number, default 0
        `threshold` that is restricted to the left
    prominence: positive number, optional
        threshold of prominence of the detected peaks (valleys)
    prominence_wlen : positive int, optional
        the `wlen` parameter of the function `scipy.signal.peak_prominences`
    edge : str or None, default "rising"
        can also be "falling", "both",
        for a flat peak, keep only the rising edge ("rising"), only the falling edge ("falling"),
        both edges ("both"), or don't detect a flat peak (None)
    kpsh : bool, default False
        keep peaks with same height even if they are closer than `mpd`
    valley : bool, default False
        if True (1), detect valleys (local minima) instead of peaks
    show : bool, default False
        if True (1), plot data in matplotlib figure
    ax : a matplotlib.axes.Axes instance, optional

    Returns
    -------
    ind : array_like
        Indeces of the peaks in `x`.

    NOTE
    ----
    The detection of valleys instead of peaks is performed internally by simply
    negating the data: ``ind_valleys = detect_peaks(-x)``.

    The function can handle NaN's.

    See this IPython Notebook [#peak]_.

    References
    ----------
    .. [#peak] http://nbviewer.ipython.org/github/demotu/BMC/blob/master/notebooks/DetectPeaks.ipynb

    Examples
    --------
    .. code-block:: python

        x = np.random.randn(100)
        x[60:81] = np.nan
        # detect all peaks and plot data
        ind = detect_peaks(x, show=True)
        print(ind)

        x = np.sin(2*np.pi*5*np.linspace(0, 1, 200)) + np.random.randn(200)/5
        # set minimum peak height = 0 and minimum peak distance = 20
        detect_peaks(x, mph=0, mpd=20, show=True)

        x = [0, 1, 0, 2, 0, 3, 0, 2, 0, 1, 0]
        # set minimum peak distance = 2
        detect_peaks(x, mpd=2, show=True)

        x = np.sin(2*np.pi*5*np.linspace(0, 1, 200)) + np.random.randn(200)/5
        # detection of valleys instead of peaks
        detect_peaks(x, mph=-1.2, mpd=20, valley=True, show=True)

        x = [0, 1, 1, 0, 1, 1, 0]
        # detect both edges
        detect_peaks(x, edge="both", show=True)

        x = [-2, 1, -2, 2, 1, 1, 3, 0]
        # set threshold = 2
        detect_peaks(x, threshold = 2, show=True)

    Version history
    ---------------
    "1.0.5":
        The sign of `mph` is inverted if parameter `valley` is True

    """
    data = deepcopy(x)
    data = np.atleast_1d(data).astype("float64")
    if data.size < 3:
        return np.array([], dtype=int)

    if valley:
        data = -data
        if mph is not None:
            mph = -mph

    # find indices of all peaks
    dx = data[1:] - data[:-1]  # equiv to np.diff()

    # handle NaN's
    indnan = np.where(np.isnan(data))[0]
    if indnan.size:
        data[indnan] = np.inf
        dx[np.where(np.isnan(dx))[0]] = np.inf

    ine, ire, ife = np.array([[], [], []], dtype=int)
    if not edge:
        ine = np.where((np.hstack((dx, 0)) < 0) & (np.hstack((0, dx)) > 0))[0]
    else:
        if edge.lower() in ["rising", "both"]:
            ire = np.where((np.hstack((dx, 0)) <= 0) & (np.hstack((0, dx)) > 0))[0]
        if edge.lower() in ["falling", "both"]:
            ife = np.where((np.hstack((dx, 0)) < 0) & (np.hstack((0, dx)) >= 0))[0]
    ind = np.unique(np.hstack((ine, ire, ife)))

    if verbose >= 1:
        print(
            f"before filtering by mpd = {mpd}, and threshold = {threshold}, ind = {ind.tolist()}"
        )
        print(
            f"additionally, left_threshold = {left_threshold}, "
            f"right_threshold = {right_threshold}, length of data = {len(data)}"
        )

    # handle NaN's
    if ind.size and indnan.size:
        # NaN's and values close to NaN's cannot be peaks
        ind = ind[
            np.in1d(
                ind, np.unique(np.hstack((indnan, indnan - 1, indnan + 1))), invert=True
            )
        ]

    if verbose >= 1:
        print(f"after handling nan values, ind = {ind.tolist()}")

    # peaks are only valid within [mpb, len(data)-mpb[
    ind = np.array([pos for pos in ind if mpd <= pos < len(data) - mpd])

    if verbose >= 1:
        print(
            f"after fitering out elements too close to border by mpd = {mpd}, ind = {ind.tolist()}"
        )

    # first and last values of data cannot be peaks
    # if ind.size and ind[0] == 0:
    #     ind = ind[1:]
    # if ind.size and ind[-1] == data.size-1:
    #     ind = ind[:-1]
    # remove peaks < minimum peak height
    if ind.size and mph is not None:
        ind = ind[data[ind] >= mph]

    if verbose >= 1:
        print(f"after filtering by mph = {mph}, ind = {ind.tolist()}")

    # remove peaks - neighbors < threshold
    _left_threshold = left_threshold if left_threshold > 0 else threshold
    _right_threshold = right_threshold if right_threshold > 0 else threshold
    if ind.size and (_left_threshold > 0 and _right_threshold > 0):
        # dx = np.min(np.vstack([data[ind]-data[ind-1], data[ind]-data[ind+1]]), axis=0)
        dx = np.max(
            np.vstack([data[ind] - data[ind + idx] for idx in range(-mpd, 0)]), axis=0
        )
        ind = np.delete(ind, np.where(dx < _left_threshold)[0])
        if verbose >= 2:
            print(f"from left, dx = {dx.tolist()}")
            print(
                f"after deleting those dx < _left_threshold = {_left_threshold}, ind = {ind.tolist()}"
            )
        dx = np.max(
            np.vstack([data[ind] - data[ind + idx] for idx in range(1, mpd + 1)]),
            axis=0,
        )
        ind = np.delete(ind, np.where(dx < _right_threshold)[0])
        if verbose >= 2:
            print(f"from right, dx = {dx.tolist()}")
            print(
                f"after deleting those dx < _right_threshold = {_right_threshold}, ind = {ind.tolist()}"
            )
    if verbose >= 1:
        print(f"after filtering by threshold, ind = {ind.tolist()}")
    # detect small peaks closer than minimum peak distance
    if ind.size and mpd > 1:
        ind = ind[np.argsort(data[ind])][::-1]  # sort ind by peak height
        idel = np.zeros(ind.size, dtype=bool)
        for i in range(ind.size):
            if not idel[i]:
                # keep peaks with the same height if kpsh is True
                idel = idel | (ind >= ind[i] - mpd) & (ind <= ind[i] + mpd) & (
                    data[ind[i]] > data[ind] if kpsh else True
                )
                idel[i] = 0  # Keep current peak
        # remove the small peaks and sort back the indices by their occurrence
        ind = np.sort(ind[~idel])

    ind = np.array(
        [
            item
            for item in ind
            if data[item] == np.max(data[item - mpd : item + mpd + 1])
        ]
    )

    if verbose >= 1:
        print(f"after filtering by mpd, ind = {ind.tolist()}")

    if prominence:
        _p = peak_prominences(data, ind, prominence_wlen)[0]
        ind = ind[np.where(_p >= prominence)[0]]
        if verbose >= 1:
            print(f"after filtering by prominence, ind = {ind.tolist()}")
            if verbose >= 2:
                print(f"with detailed prominence = {_p.tolist()}")

    return ind


def remove_spikes_naive(
    sig: np.ndarray, threshold: Real = 20, inplace: bool = True
) -> np.ndarray:
    """Remove signal spikes using a naive method.

    This is a method proposed in entry 0416 of CPSC2019.
    `spikes` here refers to abrupt large bumps with (abs) value
    larger than the given threshold,
    or nan values (read by `wfdb`).
    Do **NOT** confuse with `spikes` in paced rhythm.

    Parameters
    ----------
    sig : numpy.ndarray
        1D signal with potential spikes.
    threshold : numbers.Real, optional
        Values of `sig` that are larger than `threshold` will be removed.
    inplace : bool, optional
        Whether to modify `sig` in place or not.

    Returns
    -------
    numpy.ndarray
        Signal with `spikes` removed.

    Examples
    --------
    .. code-block:: python

        sig = np.random.randn(1000)
        pos = np.random.randint(0, 1000, 10)
        sig[pos] = 100
        sig = remove_spikes_naive(sig)
        pos = np.random.randint(0, 1000, 1)
        sig[pos] = np.nan
        sig = remove_spikes_naive(sig)

    """
    dtype = sig.dtype
    b = list(
        filter(
            lambda k: k > 0,
            np.argwhere(np.logical_or(np.abs(sig) > threshold, np.isnan(sig))).squeeze(
                -1
            ),
        )
    )
    if not inplace:
        sig = sig.copy()
    if abs(sig[0]) > threshold or np.isnan(sig[0]):
        sig[0] = 0
    for k in b:
        sig[k] = sig[k - 1]
    return sig.astype(dtype)


def butter_bandpass(
    lowcut: Real, highcut: Real, fs: Real, order: int, verbose: int = 0
) -> Tuple[np.ndarray, np.ndarray]:
    """Butterworth Bandpass Filter Design.

    Parameters
    ----------
    lowcut : numbers.Real
        Low cutoff frequency.
    highcut : numbers.Real
        High cutoff frequency.
    fs : numbers.Real
        Sampling frequency of `data`.
    order : int,
        Order of the filter.
    verbose : int, default 0
        Verbosity level for debugging.

    Returns
    -------
    b, a : numpy.ndarray
        Coefficients of numerator and denominator of the filter.

    NOTE
    ----
    According to `lowcut` and `highcut`,
    the filter type might degenerate to lowpass or highpass filter.

    References
    ----------
    1. :func:`scipy.signal.butter`
    2. https://scipy-cookbook.readthedocs.io/items/ButterworthBandpass.html

    """
    nyq = 0.5 * fs
    low = lowcut / nyq
    if low >= 1:
        raise ValueError("frequency out of range!")
    high = highcut / nyq

    if low <= 0 and high >= 1:
        raise ValueError("frequency out of range!")

    if low <= 0:
        Wn = high
        btype = "low"
    elif high >= 1:
        Wn = low
        btype = "high"
    elif lowcut == highcut:
        Wn = high
        btype = "low"
    else:
        Wn = [low, high]
        btype = "band"

    if verbose >= 1:
        print(
            f"by the setup of lowcut and highcut, the filter type falls to {btype}, with Wn = {Wn}"
        )

    b, a = butter(order, Wn, btype=btype)
    return b, a


def butter_bandpass_filter(
    data: np.ndarray,
    lowcut: Real,
    highcut: Real,
    fs: Real,
    order: int,
    btype: Optional[str] = None,
    verbose: int = 0,
) -> np.ndarray:
    """Butterworth bandpass filtering the signals.

    Apply a Butterworth bandpass filter to the signal.
    For references, see [#bp1]_ and [#bp2]_.

    Parameters
    ----------
    data : numpy.ndarray
        Signal to be filtered.
    lowcut : numbers.real
        Low cutoff frequency.
    highcut : numbers.real
        High cutoff frequency.
    fs : numbers.real
        Frequency of the signal.
    order : int
        Order of the filter.
    btype : {"lohi", "hilo"}, optional
        (special) type of the filter.
        Ignored for lowpass and highpass filters
        (as defined by `lowcut` and `highcut`).
    verbose : int, default 0
        Verbosity level for printing.

    Returns
    -------
    y : numpy.ndarray
        The filtered signal.

    References
    ----------
    .. [#bp1] https://scipy-cookbook.readthedocs.io/items/ButterworthBandpass.html
    .. [#bp2] https://dsp.stackexchange.com/questions/19084/applying-filter-in-scipy-signal-use-lfilter-or-filtfilt

    """
    dtype = data.dtype
    if btype is None:
        b, a = butter_bandpass(lowcut, highcut, fs, order=order, verbose=verbose)
        y = filtfilt(b, a, data)
        return y.astype(dtype)
    if btype.lower() == "lohi":
        b, a = butter_bandpass(0, highcut, fs, order=order, verbose=verbose)
        y = filtfilt(b, a, data)
        b, a = butter_bandpass(lowcut, fs, fs, order=order, verbose=verbose)
        y = filtfilt(b, a, y)
    elif btype.lower() == "hilo":
        b, a = butter_bandpass(lowcut, fs, fs, order=order, verbose=verbose)
        y = filtfilt(b, a, data)
        b, a = butter_bandpass(0, highcut, fs, order=order, verbose=verbose)
        y = filtfilt(b, a, y)
    else:
        raise ValueError(f"special btype `{btype}` is not supported")
    return y.astype(dtype)


def get_ampl(
    sig: np.ndarray,
    fs: Real,
    fmt: str = "lead_first",
    window: Real = 0.2,
    critical_points: Optional[Sequence] = None,
) -> Union[float, np.ndarray]:
    """Get amplitude of a signal (near critical points if given).

    Parameters
    ----------
    sig : numpy.ndarray
        (ECG) signal.
    fs : numbers.Real
        Sampling frequency of the signal
    fmt : str, default "lead_first"
        Format of the signal, can be
        "channel_last" (alias "lead_last"), or
        "channel_first" (alias "lead_first").
        Ignored if sig is 1d array (single-lead).
    window : int, default 0.2
        Window length of a window for computing amplitude, with units in seconds.
    critical_points : numpy.ndarray, optional
        Positions of critical points near which to compute amplitude,
        e.g. can be rpeaks, t peaks, etc.

    Returns
    -------
    ampl : float or numpy.ndarray
        Amplitude of the signal.

    """
    dtype = sig.dtype
    if fmt.lower() in ["channel_last", "lead_last"]:
        _sig = sig.T
    elif fmt.lower() in ["channel_first", "lead_first"]:
        _sig = sig
    else:
        raise ValueError(f"unknown format `{fmt}`")
    _window = int(round(window * fs))
    half_window = _window // 2
    _window = half_window * 2
    if critical_points is not None:
        s = np.stack(
            [
                ensure_siglen(
                    _sig[
                        ...,
                        max(0, p - half_window) : min(_sig.shape[-1], p + half_window),
                    ],
                    siglen=_window,
                    fmt="lead_first",
                )
                for p in critical_points
            ],
            axis=-1,
        ).astype(dtype)
        # the following is much slower
        # for p in critical_points:
        #     s = _sig[...,max(0,p-half_window):min(_sig.shape[-1],p+half_window)]
        #     ampl = np.max(np.array([ampl, np.max(s,axis=-1) - np.min(s,axis=-1)]), axis=0)
    else:
        s = np.stack(
            [
                _sig[..., idx * half_window : idx * half_window + _window]
                for idx in range(_sig.shape[-1] // half_window - 1)
            ],
            axis=-1,
        ).astype(dtype)
        # the following is much slower
        # for idx in range(_sig.shape[-1]//half_window-1):
        #     s = _sig[..., idx*half_window: idx*half_window+_window]
        #     ampl = np.max(np.array([ampl, np.max(s,axis=-1) - np.min(s,axis=-1)]), axis=0)
    ampl = np.max(np.max(s, axis=-2) - np.min(s, axis=-2), axis=-1)
    return ampl


def normalize(
    sig: np.ndarray,
    method: str,
    mean: Union[Real, Iterable[Real]] = 0.0,
    std: Union[Real, Iterable[Real]] = 1.0,
    sig_fmt: str = "channel_first",
    per_channel: bool = False,
) -> np.ndarray:
    """Normalize a signal.

    Perform z-score normalization on `sig`,
    to make it has fixed mean and standard deviation,
    or perform min-max normalization on `sig`,
    or normalize `sig` using `mean` and `std` via (sig - mean) / std.
    More precisely,

    .. math::

        \\begin{align*}
        \\text{Min-Max normalization:} & \\frac{sig - \\min(sig)}{\\max(sig) - \\min(sig)} \\\\
        \\text{Naive normalization:} & \\frac{sig - m}{s} \\\\
        \\text{Z-score normalization:} & \\left(\\frac{sig - mean(sig)}{std(sig)}\\right) \\cdot s + m
        \\end{align*}

    Parameters
    ----------
    sig : numpy.ndarray
        The signal to be normalized.
    method : {"naive", "min-max", "z-score"}
        Normalization method, case insensitive.
    mean : numbers.Real or array_like, default 0.0
        Mean value of the normalized signal,
        or mean values for each lead of the normalized signal.
        Useless if `method` is "min-max".
    std : numbers.Real or array_like, default 1.0
        Standard deviation of the normalized signal,
        or standard deviations for each lead of the normalized signal.
        Useless if `method` is "min-max".
    sig_fmt : str, default "channel_first"
        Format of the signal, can be of one of
        "channel_last" (alias "lead_last"), or
        "channel_first" (alias "lead_first"),
        ignored if sig is 1d array (single-lead).
    per_channel : bool, default False
        If True, normalization will be done per channel.
        Ignored if `sig` is 1d array (single-lead).

    Returns
    -------
    nm_sig : numpy.ndarray
        The normalized signal.

    NOTE
    ----
    In cases where normalization is infeasible (``std = 0``),
    only the mean value will be shifted.

    """
    assert sig.ndim in [1, 2, 3], "signal `sig` should be 1d or 2d or 3d array"
    if sig.ndim == 1 and per_channel:
        warnings.warn(
            "per-channel normalization is not supported for 1d signal, "
            "`per_channel` will be set to False",
            RuntimeWarning,
        )
        per_channel = False
    dtype = sig.dtype
    _method = method.lower()
    assert _method in [
        "z-score",
        "naive",
        "min-max",
    ], f"unknown normalization method `{method}`"
    if not per_channel:
        if sig.ndim == 2:
            assert isinstance(mean, Real) and isinstance(
                std, Real
            ), "`mean` and `std` should be real numbers in the non per-channel setting for 2d signal"
        else:  # sig.ndim == 3
            assert (isinstance(mean, Real) or np.shape(mean) == (sig.shape[0],)) and (
                isinstance(std, Real) or np.shape(std) == (sig.shape[0],)
            ), (
                f"`mean` and `std` should be real numbers or have shape ({sig.shape[0]},) "
                "in the non per-channel setting for 3d signal"
            )
    if isinstance(std, Real):
        assert std > 0, "standard deviation should be positive"
    else:
        assert (np.array(std) > 0).all(), "standard deviations should all be positive"
    assert sig_fmt.lower() in [
        "channel_first",
        "lead_first",
        "channel_last",
        "lead_last",
    ], f"format `{sig_fmt}` of the signal not supported!"

    if isinstance(mean, Iterable):
        assert sig.ndim in [2, 3], "`mean` should be a real number for 1d signal"
        if sig.ndim == 2:
            assert np.shape(mean) in [
                (sig.shape[0],),
                (sig.shape[-1],),
            ], f"shape of `mean` = {np.shape(mean)} not compatible with the `sig` = {np.shape(sig)}"
            if sig_fmt.lower() in [
                "channel_first",
                "lead_first",
            ]:
                _mean = np.array(mean, dtype=dtype)[..., np.newaxis]
            else:
                _mean = np.array(mean, dtype=dtype)[np.newaxis, ...]
        else:  # sig.ndim == 3
            if sig_fmt.lower() in [
                "channel_first",
                "lead_first",
            ]:
                if np.shape(mean) == (sig.shape[0],):
                    _mean = np.array(mean, dtype=dtype)[..., np.newaxis, np.newaxis]
                elif np.shape(mean) == (sig.shape[1],):
                    _mean = np.repeat(
                        np.array(mean, dtype=dtype)[np.newaxis, ..., np.newaxis],
                        sig.shape[0],
                        axis=0,
                    )
                elif np.shape(mean) == sig.shape[:2]:
                    _mean = np.array(mean, dtype=dtype)[..., np.newaxis]
                else:
                    raise AssertionError(
                        f"shape of `mean` = {np.shape(mean)} not compatible with the `sig` = {np.shape(sig)}"
                    )
            else:  # "channel_last" or "lead_last"
                if np.shape(mean) == (sig.shape[0],):
                    _mean = np.array(mean, dtype=dtype)[..., np.newaxis, np.newaxis]
                elif np.shape(mean) == (sig.shape[-1],):
                    _mean = np.repeat(
                        np.array(mean, dtype=dtype)[np.newaxis, np.newaxis, ...],
                        sig.shape[0],
                        axis=0,
                    )
                elif np.shape(mean) == (sig.shape[0], sig.shape[-1]):
                    _mean = np.expand_dims(np.array(mean, dtype=dtype), axis=1)
                else:
                    raise AssertionError(
                        f"shape of `mean` = {np.shape(mean)} not compatible with the `sig` = {np.shape(sig)}"
                    )
    else:
        _mean = mean
    if isinstance(std, Iterable):
        assert sig.ndim in [2, 3], "`std` should be a real number for 1d signal"
        if sig.ndim == 2:
            assert np.shape(std) in [
                (sig.shape[0],),
                (sig.shape[-1],),
            ], f"shape of `std` = {np.shape(std)} not compatible with the `sig` = {np.shape(sig)}"
            if sig_fmt.lower() in [
                "channel_first",
                "lead_first",
            ]:
                _std = np.array(std, dtype=dtype)[..., np.newaxis]
            else:
                _std = np.array(std, dtype=dtype)[np.newaxis, ...]
        else:  # sig.ndim == 3
            if sig_fmt.lower() in [
                "channel_first",
                "lead_first",
            ]:
                if np.shape(std) == (sig.shape[0],):
                    _std = np.array(std, dtype=dtype)[..., np.newaxis, np.newaxis]
                elif np.shape(std) == (sig.shape[1],):
                    _std = np.repeat(
                        np.array(std, dtype=dtype)[np.newaxis, ..., np.newaxis],
                        sig.shape[0],
                        axis=0,
                    )
                elif np.shape(std) == sig.shape[:2]:
                    _std = np.array(std, dtype=dtype)[..., np.newaxis]
                else:
                    raise AssertionError(
                        f"shape of `std` = {np.shape(std)} not compatible with the `sig` = {np.shape(sig)}"
                    )
            else:  # "channel_last" or "lead_last"
                if np.shape(std) == (sig.shape[0],):
                    _std = np.array(std, dtype=dtype)[..., np.newaxis, np.newaxis]
                elif np.shape(std) == (sig.shape[-1],):
                    _std = np.repeat(
                        np.array(std, dtype=dtype)[np.newaxis, np.newaxis, ...],
                        sig.shape[0],
                        axis=0,
                    )
                elif np.shape(std) == (sig.shape[0], sig.shape[-1]):
                    _std = np.expand_dims(np.array(std, dtype=dtype), axis=1)
                else:
                    raise AssertionError(
                        f"shape of `std` = {np.shape(std)} not compatible with the `sig` = {np.shape(sig)}"
                    )
    else:
        _std = std

    if _method == "naive":
        nm_sig = (sig - _mean) / _std
        return nm_sig.astype(dtype)

    eps = 1e-7  # to avoid dividing by zero
    if sig.ndim == 3:  # the first dimension is the batch dimension
        if not per_channel:
            options = dict(axis=(1, 2), keepdims=True)
        elif sig_fmt.lower() in [
            "channel_first",
            "lead_first",
        ]:
            options = dict(axis=2, keepdims=True)
        else:
            options = dict(axis=1, keepdims=True)
    else:
        if not per_channel:
            options = dict(axis=None)
        elif sig_fmt.lower() in [
            "channel_first",
            "lead_first",
        ]:
            options = dict(axis=1, keepdims=True)
        else:
            options = dict(axis=0, keepdims=True)

    if _method == "z-score":
        nm_sig = (
            (sig - np.mean(sig, dtype=dtype, **options))
            / (np.std(sig, dtype=dtype, **options) + eps)
        ) * _std + _mean
    elif _method == "min-max":
        nm_sig = (sig - np.amin(sig, **options)) / (
            np.amax(sig, **options) - np.amin(sig, **options) + eps
        )
    return nm_sig.astype(dtype)
